# Blackbox tests for smbcacls
#
# Copyright (C) Noel Power noel.power@suse.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from samba.tests.blackbox.smbcacls import SmbCaclsBlockboxTestBase
from samba.tests import BlackboxProcessError
import os

class SaveRestoreSmbCaclsTests(SmbCaclsBlockboxTestBase):

    def setUp(self):
        super().setUp()

        # create toplevel testdir structure with desired ACL(s)
        #
        #  +-tar_test_dir/    (OI)(CI)(I)(F)
        #  +-oi_dir/        (OI)(CI)(I)(F)
        #  | +-file.1            (I)(F)
        #  | +-nested/      (OI)(CI)(I)(F)
        #  |   +-file.2          (I)(F)
        #  |   +-nested_again/     (OI)(CI)(I)(F)
        #  |     +-file.3          (I)(F)

        self.toplevel = self.create_remote_test_file("tar_test_dir/file-0")
        self.f1 = self.create_remote_test_file("tar_test_dir/oi_dir/file-1")
        self.f2 = self.create_remote_test_file("tar_test_dir/oi_dir/nested/file-2")
        self.f3 = self.create_remote_test_file("tar_test_dir/oi_dir/nested/nested_again/file-3")
        self.tar_dir = os.path.split(self.toplevel)[0]
        self.oi_dir = os.path.split(self.f1)[0]
        self.nested_dir = os.path.split(self.f2)[0]
        self.nested_again_dir = os.path.split(self.f3)[0]

        dir_acl_str = "ACL:%s:ALLOWED/OI|CI/FULL" % self.user
        inherited_dir_acl_str = "ACL:%s:ALLOWED/OI|CI|I/FULL" % self.user
        file_acl_str = "ACL:%s:ALLOWED/I/FULL" % self.user

        self.smb_cacls(["--modify", dir_acl_str, self.tar_dir])
        self.smb_cacls(["--modify", inherited_dir_acl_str, self.oi_dir])
        self.smb_cacls(["--modify", inherited_dir_acl_str, self.nested_dir])
        self.smb_cacls(["--modify", inherited_dir_acl_str, self.nested_again_dir])
        self.smb_cacls(["--modify", file_acl_str, self.f1])
        self.smb_cacls(["--modify", file_acl_str, self.f2])
        self.smb_cacls(["--modify", file_acl_str, self.f3])

    def tearDown(self):
        # tmp is the default share which has an existing testdir smbcacls
        # we need to be prepared to deal with a 'custom' share (which also
        # would have an existing testdir)
        if self.share != "tmp":
            self.dirpath = os.path.join(os.environ["LOCAL_PATH"],self.share)
            self.dirpath = os.path.join(self.dirpath,self.testdir)
        super().tearDown()

    def test_simple_save_dir(self):
        try:
            # simple test to just store dacl of directory
            with self.mktemp() as tmpfile:
                out = self.smb_cacls(["--save", tmpfile,
                            self.oi_dir])
                with open(tmpfile, 'rb') as infile:
                    contents = infile.read().decode('utf16')
                    lines = contents.splitlines()
                    # should be 2 lines
                    self.assertEqual(len(lines), 2)
                    # first line should be the path
                    self.assertEqual(self.oi_dir.replace('/','\\'), lines[0])

        except BlackboxProcessError as e:
            self.fail(str(e))

    def test_simple_save_dir_r(self):
        try:
            # simple test to just store dacl of directory (recursively)
            with self.mktemp() as tmpfile:
                out = self.smb_cacls(["--recurse", "--save", tmpfile,
                            self.oi_dir])
                with open(tmpfile, 'rb') as infile:
                    contents = infile.read().decode('utf16')
                    print("contents = %s" % contents)
                    lines = contents.splitlines()
                    # should be 12 lines
                    self.assertEqual(len(lines), 12)
                    paths = [
                            self.oi_dir.replace('/','\\'),
                            self.f1.replace('/','\\'),
                            self.nested_dir.replace('/','\\'),
                            self.f2.replace('/','\\'),
                            self.nested_again_dir.replace('/','\\'),
                            self.f3.replace('/','\\')
                            ]
                    i = 0
                    for line in lines:
                        if not i % 2:
                            paths.remove(line)
                        i = i + 1
                    self.assertEqual(0, len(paths))

        except BlackboxProcessError as e:
            self.fail(str(e))

    def test_simple_restore_dir(self):
        try:
            # simple test to just store dacl of directory
            orig_saved = None
            modified = None
            restored = None
            with self.mktemp() as tmpfile:
                self.smb_cacls(["--save", tmpfile,
                            self.oi_dir])
                with open(tmpfile, 'rb') as infile:
                    orig_saved = infile.read()

            # modify directory structure
            dir_add_acl_str = "ACL:%s:ALLOWED/OI|CI/READ" % self.user
            self.smb_cacls(["--propagate-inheritance", "--add",
                           dir_add_acl_str, self.oi_dir])

            # save modified directory dacls to file
            with self.mktemp() as tmpfile:
                self.smb_cacls(["--save", tmpfile,
                            self.oi_dir])
                with open(tmpfile, 'rb') as infile:
                    modified = infile.read()

            # compare orig and unmodified dacls
            # they shouldn't match
            self.assertNotEqual(orig_saved.decode('utf16'), modified.decode('utf16'))
            # restore original dacls from file
            with self.mktemp() as tmpfile:
                with open(tmpfile, 'wb') as outfile:
                    outfile.write(orig_saved)
                    outfile.close()
                    out = self.smb_cacls([".", "--restore", tmpfile])

            # save newly restored dacls to file
            with self.mktemp() as tmpfile:
                self.smb_cacls(["--save", tmpfile,
                            self.oi_dir])
                with open(tmpfile, 'rb') as infile:
                    restored = infile.read()

            # after restoring the dalcs, orig unmodified dacls should match
            # restored dacls
            self.assertEqual(orig_saved.decode('utf16'), restored.decode('utf16'))

        except BlackboxProcessError as e:
            self.fail(str(e))

    def test_simple_restore_dir_r(self):
        try:
            # simple test to just store dacl(s) of directory recursively
            orig_saved = None
            modified = None
            restored = None
            with self.mktemp() as tmpfile:
                self.smb_cacls(["--recurse", "--save", tmpfile,
                            self.oi_dir])
                with open(tmpfile, 'rb') as infile:
                    orig_saved = infile.read()

            # modify directory's dacls recursively
            dir_add_acl_str = "ACL:%s:ALLOWED/OI|CI/READ" % self.user
            self.smb_cacls(["--propagate-inheritance", "--add",
                           dir_add_acl_str, self.oi_dir])

            # save modified directories dacls recursively
            with self.mktemp() as tmpfile:
                self.smb_cacls(["--recurse", "--save", tmpfile,
                            self.oi_dir])
                with open(tmpfile, 'rb') as infile:
                    modified = infile.read()

            # the unmodified stringified dacls shouldn't match
            # modified
            self.assertNotEqual(orig_saved.decode('utf16'), modified.decode('utf16'))
            # restore original dacls from file
            with self.mktemp() as tmpfile:
                with open(tmpfile, 'wb') as outfile:
                    outfile.write(orig_saved)
                    outfile.close()
                    out = self.smb_cacls([".", "--restore", tmpfile])

            with self.mktemp() as tmpfile:
                out = self.smb_cacls(["--recurse", "--save", tmpfile,
                            self.oi_dir])
                with open(tmpfile, 'rb') as infile:
                    restored = infile.read()
            # after restoring the dalcs orig unmodified dacls should match
            # current dacls
            self.assertEqual(orig_saved.decode('utf16'), restored.decode('utf16'))
        except BlackboxProcessError as e:
            self.fail(str(e))



# vgp_files_ext samba gpo policy
# Copyright (C) David Mulder <dmulder@suse.com> 2020
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, pwd, grp
from samba.gp.gpclass import gp_xml_ext, check_safe_path, gp_file_applier
from tempfile import NamedTemporaryFile
from shutil import copyfile, move
from samba.gp.util.logging import log

def calc_mode(entry):
    mode = 0o000
    for permissions in entry.findall('permissions'):
        ptype = permissions.get('type')
        if ptype == 'user':
            if permissions.find('read') is not None:
                mode |= 0o400
            if permissions.find('write') is not None:
                mode |= 0o200
            if permissions.find('execute') is not None:
                mode |= 0o100
        elif ptype == 'group':
            if permissions.find('read') is not None:
                mode |= 0o040
            if permissions.find('write') is not None:
                mode |= 0o020
            if permissions.find('execute') is not None:
                mode |= 0o010
        elif ptype == 'other':
            if permissions.find('read') is not None:
                mode |= 0o004
            if permissions.find('write') is not None:
                mode |= 0o002
            if permissions.find('execute') is not None:
                mode |= 0o001
    return mode

def stat_from_mode(mode):
    stat = '-'
    for i in range(6, -1, -3):
        mask = {0o4: 'r', 0o2: 'w', 0o1: 'x'}
        for x in mask.keys():
            if mode & (x << i):
                stat += mask[x]
            else:
                stat += '-'
    return stat

def source_file_change(fname):
    if os.path.exists(fname):
        return b'%d' % os.stat(fname).st_ctime

class vgp_files_ext(gp_xml_ext, gp_file_applier):
    def __str__(self):
        return 'VGP/Unix Settings/Files'

    def process_group_policy(self, deleted_gpo_list, changed_gpo_list):
        for guid, settings in deleted_gpo_list:
            if str(self) in settings:
                for attribute, _ in settings[str(self)].items():
                    self.unapply(guid, attribute, attribute)

        for gpo in changed_gpo_list:
            if gpo.file_sys_path:
                xml = 'MACHINE/VGP/VTLA/Unix/Files/manifest.xml'
                path = os.path.join(gpo.file_sys_path, xml)
                xml_conf = self.parse(path)
                if not xml_conf:
                    continue
                policy = xml_conf.find('policysetting')
                data = policy.find('data')
                for entry in data.findall('file_properties'):
                    local_path = self.lp.cache_path('gpo_cache')
                    source = entry.find('source').text
                    source_file = os.path.join(local_path,
                        os.path.dirname(check_safe_path(path)).upper(),
                                        source.upper())
                    if not os.path.exists(source_file):
                        log.warn('Source file does not exist', source_file)
                        continue
                    target = entry.find('target').text
                    user = entry.find('user').text
                    group = entry.find('group').text
                    mode = calc_mode(entry)

                    # The attribute is simply the target file.
                    attribute = target
                    # The value hash is generated from the source file last
                    # change stamp, the user, the group, and the mode, ensuring
                    # any changes to this GPO will cause the file to be
                    # rewritten.
                    value_hash = self.generate_value_hash(
                            source_file_change(source_file),
                            user, group, b'%d' % mode)
                    def applier_func(source_file, target, user, group, mode):
                        with NamedTemporaryFile(dir=os.path.dirname(target),
                                                delete=False) as f:
                            copyfile(source_file, f.name)
                            os.chown(f.name, pwd.getpwnam(user).pw_uid,
                                     grp.getgrnam(group).gr_gid)
                            os.chmod(f.name, mode)
                            move(f.name, target)
                        return [target]
                    self.apply(gpo.name, attribute, value_hash, applier_func,
                               source_file, target, user, group, mode)

    def rsop(self, gpo):
        output = {}
        xml = 'MACHINE/VGP/VTLA/Unix/Files/manifest.xml'
        if gpo.file_sys_path:
            path = os.path.join(gpo.file_sys_path, xml)
            xml_conf = self.parse(path)
            if not xml_conf:
                return output
            policy = xml_conf.find('policysetting')
            data = policy.find('data')
            for entry in data.findall('file_properties'):
                source = entry.find('source').text
                target = entry.find('target').text
                user = entry.find('user').text
                group = entry.find('group').text
                mode = calc_mode(entry)
                p = '%s\t%s\t%s\t%s -> %s' % \
                    (stat_from_mode(mode), user, group, target, source)
                if str(self) not in output.keys():
                    output[str(self)] = []
                output[str(self)].append(p)
        return output



# Blackbox tests for the smbcontrol fault injection commands
#
# Copyright (C) Andrew Bartlett <abartlet@samba.org> 2018
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# As the test terminates and sleeps samba processes these tests need to run
# in the preforkrestartdc test environment to prevent them impacting other
# tests.
#
import time
from samba.tests import BlackboxTestCase, BlackboxProcessError
from samba.messaging import Messaging

COMMAND = "bin/smbcontrol"
PING = "ping"


class SmbcontrolProcessBlockboxTests(BlackboxTestCase):

    def setUp(self):
        super().setUp()
        lp_ctx = self.get_loadparm()
        self.msg_ctx = Messaging(lp_ctx=lp_ctx)

    def get_process_data(self):
        services = self.msg_ctx.irpc_all_servers()

        processes = []
        for service in services:
            for id in service.ids:
                processes.append((service.name, id.pid))
        return processes

    def get_process(self, name):
        processes = self.get_process_data()
        for pname, pid in processes:
            if name == pname:
                return pid
        return None

    def test_inject_fault(self):
        INJECT = "inject"
        FAULT = "segv"
        #
        # Note that this process name needs to be different to the one used
        # in the sleep test to avoid a race condition
        #
        pid = self.get_process("rpc_server")

        #
        # Ensure we can ping the process before injecting a fault.
        #
        try:
            self.check_run("%s %s %s" % (COMMAND, pid, PING),
                           msg="trying to ping rpc_server")
        except BlackboxProcessError as e:
            self.fail("Unable to ping rpc_server process")

        #
        # Now inject a fault.
        #
        try:
            self.check_run("%s %s %s %s" % (COMMAND, pid, INJECT, FAULT),
                           msg="injecting fault into rpc_server")
        except BlackboxProcessError as e:
            print(e)
            self.fail("Unable to inject a fault into the rpc_server process")

        #
        # The process should have died, so we should not be able to ping it
        #
        try:
            self.check_run("%s %s %s" % (COMMAND, pid, PING),
                           msg="trying to ping rpc_server")
            self.fail("Could ping rpc_server process")
        except BlackboxProcessError as e:
            pass

    def test_sleep(self):
        SLEEP = "sleep"  # smbcontrol sleep command
        DURATION = 5     # duration to sleep server for
        DELTA = 1        # permitted error for the sleep duration

        #
        # Note that this process name needs to be different to the one used
        # in the inject fault test to avoid a race condition
        #
        pid = self.get_process("ldap_server")
        #
        # Ensure we can ping the process before getting it to sleep
        #
        try:
            self.check_run("%s %s %s" % (COMMAND, pid, PING),
                           msg="trying to ping rpc_server")
        except BlackboxProcessError as e:
            self.fail("Unable to ping rpc_server process")

        #
        # Now ask it to sleep
        #
        start = time.time()
        try:
            self.check_run("%s %s %s %s" % (COMMAND, pid, SLEEP, DURATION),
                           msg="putting rpc_server to sleep for %d" % DURATION)
        except BlackboxProcessError as e:
            print(e)
            self.fail("Failed to get rpc_server to sleep for %d" % DURATION)

        #
        # The process should be sleeping and not respond until it wakes
        #
        try:
            self.check_run("%s %s %s" % (COMMAND, pid, PING),
                           msg="trying to ping rpc_server")
            end = time.time()
            duration = end - start
            self.assertGreater(duration + DELTA, DURATION)
        except BlackboxProcessError as e:
            self.fail("Unable to ping rpc_server process")



# Simple subunit testrunner for python

# NOTE: DO NOT USE THIS MODULE FOR NEW CODE.
#
# Instead, use the standard subunit runner - e.g. "python -m subunit.run
# YOURMODULE".
#
# This wrapper will be removed once all tests can be run
# without it. At the moment there are various tests which still
# get e.g. credentials passed via command-line options to this
# script.

# Copyright (C) Jelmer Vernooij <jelmer@samba.org> 2007-2014
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# make sure the script dies immediately when hitting control-C,
# rather than raising KeyboardInterrupt. As we do all database
# operations using transactions, this is safe.
import signal
signal.signal(signal.SIGINT, signal.SIG_DFL)

import optparse
import sys

from samba.subunit.run import TestProgram as BaseTestProgram


class SubunitOptions(optparse.OptionGroup):
    """Command line options for subunit test runners."""

    def __init__(self, parser):
        optparse.OptionGroup.__init__(self, parser, "Subunit Options")
        self.add_option('-l', '--list', dest='listtests', default=False,
                        help='List tests rather than running them.',
                        action="store_true")
        self.add_option('--load-list', dest='load_list', default=None,
                        help='Specify a filename containing the test ids to use.')


class TestProgram(BaseTestProgram):

    def __init__(self, module=None, args=None, opts=None):
        if args is None:
            args = []
        if getattr(opts, "listtests", False):
            args.insert(0, "--list")
        if getattr(opts, 'load_list', None):
            args.insert(0, "--load-list=%s" % opts.load_list)
        argv = [sys.argv[0]] + args
        super().__init__(module=module, argv=argv)



# Unix SMB/CIFS implementation.
# Copyright (C) Jelmer Vernooij <jelmer@samba.org> 2007-2012
#
# Tests for documentation.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for presence of documentation."""

import samba
import samba.tests

import os
import subprocess
import xml.etree.ElementTree as ET
import multiprocessing
import concurrent.futures
import tempfile

class TestCase(samba.tests.TestCaseInTempDir):

    def _format_message(self, parameters, message):
        parameters = list(parameters)
        parameters = list(map(str, parameters))
        parameters.sort()
        return message + '\n\n    %s' % ('\n    '.join(parameters))

def get_max_worker_count():
    cpu_count = multiprocessing.cpu_count()

    # Always run two processes in parallel
    if cpu_count < 2:
        return 2

    max_workers = int(cpu_count / 2)
    if max_workers < 2:
        return 2

    return max_workers

def check_or_set_smbconf_default(cmdline, topdir, param, default_param):
    p = subprocess.Popen(cmdline,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE,
                         cwd=topdir).communicate()
    result = p[0].decode().upper().strip()
    if result != default_param.upper():
        if not (result == "" and default_param == '""'):
            return result, param, default_param

    return None

def set_smbconf_arbitrary(cmdline, topdir, param, param_type, value_to_use):
    p = subprocess.Popen(cmdline,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE,
                         cwd=topdir).communicate()
    result = p[0].decode().upper().strip()
    if result != value_to_use.upper():
        # currently no way to distinguish command lists
        if param_type == 'list':
            if ", ".join(result.split()) == value_to_use.upper():
                return None

    # currently no way to identify octal
    if param_type == 'integer':
        try:
            if int(value_to_use, 8) == int(p[0].strip(), 8):
                return None
        except:
            pass

        return result, param, value_to_use

    return None

def set_smbconf_arbitrary_opposite(cmdline, topdir, tempdir, section, param,
                                   param_type, opposite_value, value_to_use):
    g = tempfile.NamedTemporaryFile(mode='w', dir=tempdir, delete=False)
    try:
        towrite = section + "\n"
        towrite += param + " = " + opposite_value
        g.write(towrite)
    finally:
        g.close()

    p = subprocess.Popen(cmdline + ["-s", g.name],
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE,
                         cwd=topdir).communicate()
    os.unlink(g.name)

    # testparm doesn't display a value if they are equivalent
    if (value_to_use.lower() != opposite_value.lower()):
        for line in p[0].decode().splitlines():
            if not line.strip().startswith(param):
                return None

            value_found = line.split("=")[1].upper().strip()
            if value_found != value_to_use.upper():
                # currently no way to distinguish command lists
                if param_type == 'list':
                    if ", ".join(value_found.split()) == value_to_use.upper():
                        return None

                # currently no way to identify octal
                if param_type == 'integer':
                    try:
                        if int(value_to_use, 8) == int(value_found, 8):
                            continue
                    except:
                        pass

                    return param, value_to_use, value_found

    return None

def get_documented_parameters(sourcedir):
    path = os.path.join(sourcedir, "bin", "default", "docs-xml", "smbdotconf")
    if not os.path.exists(os.path.join(path, "parameters.all.xml")):
        raise Exception("Unable to find parameters.all.xml")
    try:
        p = open(os.path.join(path, "parameters.all.xml"), 'r')
    except IOError as e:
        raise Exception("Error opening parameters file")
    out = p.read()

    root = ET.fromstring(out)
    for parameter in root:
        name = parameter.attrib.get('name')
        if parameter.attrib.get('removed') == "1":
            continue
        yield name
        syn = parameter.findall('synonym')
        if syn is not None:
            for sy in syn:
                yield sy.text
    p.close()


def get_documented_tuples(sourcedir, omit_no_default=True):
    path = os.path.join(sourcedir, "bin", "default", "docs-xml", "smbdotconf")
    if not os.path.exists(os.path.join(path, "parameters.all.xml")):
        raise Exception("Unable to find parameters.all.xml")
    try:
        p = open(os.path.join(path, "parameters.all.xml"), 'r')
    except IOError as e:
        raise Exception("Error opening parameters file")
    out = p.read()

    root = ET.fromstring(out)
    for parameter in root:
        name = parameter.attrib.get("name")
        param_type = parameter.attrib.get("type")
        if parameter.attrib.get('removed') == "1":
            continue
        values = parameter.findall("value")
        defaults = []
        for value in values:
            if value.attrib.get("type") == "default":
                defaults.append(value)

        default_text = None
        if len(defaults) == 0:
            if omit_no_default:
                continue
        elif len(defaults) > 1:
            raise Exception("More than one default found for parameter %s" % name)
        else:
            default_text = defaults[0].text

        if default_text is None:
            default_text = ""
        context = parameter.attrib.get("context")
        yield name, default_text, context, param_type
    p.close()


class SmbDotConfTests(TestCase):

    # defines the cases where the defaults may differ from the documentation
    #
    # Please pass the default via waf rather than adding to this
    # list if at all possible.
    special_cases = set([
        'log level',
        'path',
        'panic action',
        'homedir map',
        'NIS homedir',
        'server string',
        'netbios name',
        'socket options',
        'ctdbd socket',
        'printing',
        'printcap name',
        'queueresume command',
        'queuepause command',
        'lpresume command',
        'lppause command',
        'lprm command',
        'lpq command',
        'print command',
        'template homedir',
        'max open files',
        'include system krb5 conf',
        'smbd max async dosmode',
        'dns hostname',
    ])

    def setUp(self):
        super().setUp()
        # create a minimal smb.conf file for testparm
        self.smbconf = os.path.join(self.tempdir, "paramtestsmb.conf")
        f = open(self.smbconf, 'w')
        try:
            f.write("""
[test]
   path = /
""")
        finally:
            f.close()

        self.blankconf = os.path.join(self.tempdir, "emptytestsmb.conf")
        f = open(self.blankconf, 'w')
        try:
            f.write("")
        finally:
            f.close()

        self.topdir = os.path.abspath(samba.tests.source_tree_topdir())

        try:
            self.documented = set(get_documented_parameters(self.topdir))
        except:
            self.fail("Unable to load documented parameters")

        try:
            self.defaults = set(get_documented_tuples(self.topdir))
        except:
            self.fail("Unable to load parameters")

        try:
            self.defaults_all = set(get_documented_tuples(self.topdir, False))
        except:
            self.fail("Unable to load parameters")

    def tearDown(self):
        super().tearDown()
        os.unlink(self.smbconf)
        os.unlink(self.blankconf)

    def test_default_s3(self):
        self._test_default(['bin/testparm'])
        self._set_defaults(['bin/testparm'])

        # registry shares appears to need sudo
        self._set_arbitrary(['bin/testparm'],
            exceptions = ['client lanman auth',
                          'client plaintext auth',
                          'registry shares',
                          'rpc server dynamic port range',
                          'name resolve order',
                          'clustering'])
        self._test_empty(['bin/testparm'])

    def test_default_s4(self):
        self._test_default(['bin/samba-tool', 'testparm'])
        self._set_defaults(['bin/samba-tool', 'testparm'])
        self._set_arbitrary(['bin/samba-tool', 'testparm'],
                            exceptions=['rpc server dynamic port range',
                                        'name resolve order'])
        self._test_empty(['bin/samba-tool', 'testparm'])

    def _test_default(self, program):

        if program[0] == 'bin/samba-tool' and os.getenv("PYTHON", None):
            program = [os.environ["PYTHON"]] + program

        failset = set()

        mp_context = multiprocessing.get_context('fork')
        with concurrent.futures.ProcessPoolExecutor(
                max_workers=get_max_worker_count(),
                mp_context=mp_context) as executor:
            result_futures = []

            for tuples in self.defaults:
                param, default, context, param_type = tuples

                if param in self.special_cases:
                    continue
                # bad, bad parametric options - we don't have their default values
                if ':' in param:
                    continue
                section = None
                if context == "G":
                    section = "global"
                elif context == "S":
                    section = "test"
                else:
                    self.fail("%s has no valid context" % param)

                program_arg1 = ["--configfile=%s" % (self.smbconf)]
                if (program[0] == 'bin/testparm'):
                    program_arg1 = ["--suppress-prompt", self.smbconf]

                cmdline = program + program_arg1 + [
                                     "--section-name",
                                     section,
                                     "--parameter-name",
                                     param]

                future = executor.submit(check_or_set_smbconf_default, cmdline, self.topdir, param, default)
                result_futures.append(future)

            for f in concurrent.futures.as_completed(result_futures):
                if f.result():
                    result, param, default_param = f.result()

                    doc_triple = "%s\n      Expected: %s" % (param, default_param)
                    failset.add("%s\n      Got: %s" % (doc_triple, result))

        if len(failset) > 0:
            self.fail(self._format_message(failset,
                                           "Parameters that do not have matching defaults:"))

    def _set_defaults(self, program):

        if program[0] == 'bin/samba-tool' and os.getenv("PYTHON", None):
            program = [os.environ["PYTHON"]] + program

        failset = set()

        mp_context = multiprocessing.get_context('fork')
        with concurrent.futures.ProcessPoolExecutor(
                max_workers=get_max_worker_count(),
                mp_context=mp_context) as executor:
            result_futures = []

            for tuples in self.defaults:
                param, default, context, param_type = tuples

                exceptions = set([
                    'printing',
                    'smbd max async dosmode',
                ])

                if param in exceptions:
                    continue

                section = None
                if context == "G":
                    section = "global"
                elif context == "S":
                    section = "test"
                else:
                    self.fail("%s has no valid context" % param)

                program_arg1 = ["--configfile=%s" % (self.smbconf)]
                if (program[0] == 'bin/testparm'):
                    program_arg1 = ["--suppress-prompt", self.smbconf]

                cmdline = program + program_arg1 + [
                                     "--section-name",
                                     section,
                                     "--parameter-name",
                                     param,
                                     "--option",
                                     "%s = %s" % (param, default)]
                future = executor.submit(check_or_set_smbconf_default, cmdline, self.topdir, param, default)
                result_futures.append(future)

            for f in concurrent.futures.as_completed(result_futures):
                if f.result():
                    result, param, default_param = f.result()

                    doc_triple = "%s\n      Expected: %s" % (param, default)
                    failset.add("%s\n      Got: %s" % (doc_triple, result))

        if len(failset) > 0:
            self.fail(self._format_message(failset,
                                           "Parameters that do not have matching defaults:"))

    def _set_arbitrary(self, program, exceptions=None):

        if program[0] == 'bin/samba-tool' and os.getenv("PYTHON", None):
            program = [os.environ["PYTHON"]] + program

        arbitrary = {'string': 'string', 'boolean': 'yes', 'integer': '5',
                     'boolean-rev': 'yes',
                     'cmdlist': 'a b c',
                     'bytes': '10',
                     'octal': '0123',
                     'ustring': 'ustring',
                     'enum': '', 'boolean-auto': '', 'char': 'a', 'list': 'a, b, c'}
        opposite_arbitrary = {'string': 'string2', 'boolean': 'no', 'integer': '6',
                              'boolean-rev': 'no',
                              'cmdlist': 'd e f',
                              'bytes': '11',
                              'octal': '0567',
                              'ustring': 'ustring2',
                              'enum': '', 'boolean-auto': '', 'char': 'b', 'list': 'd, e, f'}

        failset = set()

        mp_context = multiprocessing.get_context('fork')
        with concurrent.futures.ProcessPoolExecutor(
                max_workers=get_max_worker_count(),
                mp_context=mp_context) as executor:
            result_futures1 = []
            result_futures2 = []

            for tuples in self.defaults_all:
                param, default, context, param_type = tuples

                if param in ['printing', 'copy', 'include', 'log level']:
                    continue

                # currently no easy way to set an arbitrary value for these
                if param_type in ['enum', 'boolean-auto']:
                    continue

                if exceptions is not None:
                    if param in exceptions:
                        continue

                section = None
                if context == "G":
                    section = "global"
                elif context == "S":
                    section = "test"
                else:
                    self.fail("%s has no valid context" % param)

                value_to_use = arbitrary.get(param_type)
                if value_to_use is None:
                    self.fail("%s has an invalid type" % param)

                program_arg1 = ["--configfile=%s" % (self.smbconf)]
                if (program[0] == 'bin/testparm'):
                    program_arg1 = ["--suppress-prompt", self.smbconf]

                cmdline = program + program_arg1 + [
                                     "--section-name",
                                     section,
                                     "--parameter-name",
                                     param,
                                     "--option",
                                     "%s = %s" % (param, value_to_use)]

                future = executor.submit(set_smbconf_arbitrary, cmdline, self.topdir, param, param_type, value_to_use)
                result_futures1.append(future)

                opposite_value = opposite_arbitrary.get(param_type)

                cmdline = program + ["--suppress-prompt",
                                     "--option",
                                     "%s = %s" % (param, value_to_use)]

                future = executor.submit(set_smbconf_arbitrary_opposite, cmdline, self.topdir, self.tempdir,
                                         section, param, param_type, opposite_value, value_to_use)
                result_futures2.append(future)

            for f in concurrent.futures.as_completed(result_futures1):
                if f.result():
                    result, param, value_to_use = f.result()

                    doc_triple = "%s\n      Expected: %s" % (param, value_to_use)
                    failset.add("%s\n      Got: %s" % (doc_triple, result))

            for f in concurrent.futures.as_completed(result_futures2):
                if f.result():
                    param, value_to_use, value_found = f.result()

                    doc_triple = "%s\n      Expected: %s" % (param, value_to_use)
                    failset.add("%s\n      Got: %s" % (doc_triple, value_found))

        if len(failset) > 0:
            self.fail(self._format_message(failset,
                                           "Parameters that were unexpectedly not set:"))

    def _test_empty(self, program):

        if program[0] == 'bin/samba-tool' and os.getenv("PYTHON", None):
            program = [os.environ["PYTHON"]] + program

        program_arg1 = ["--configfile=%s" % (self.blankconf), "--suppress-prompt"]
        if (program[0] == 'bin/testparm'):
            program_arg1 = ["--suppress-prompt", self.blankconf]

        print(program + program_arg1)
        p = subprocess.Popen(program + program_arg1,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             cwd=self.topdir).communicate()
        output = ""

        for line in p[0].decode().splitlines():
            if line.strip().startswith('#'):
                continue
            if line.strip().startswith("idmap config *"):
                continue
            output += line.strip().lower() + '\n'

        if output.strip() != '[global]' and output.strip() != '[globals]':
            self.fail("Testparm returned unexpected output on an empty smb.conf.")



# Unix SMB/CIFS implementation.
# Copyright (C) William Brown <william@blackhats.net.au> 2018
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import ldb
from samba.tests.samba_tool.base import SambaToolCmdTest


class ForestCmdTestCase(SambaToolCmdTest):
    """Tests for samba-tool dsacl subcommands"""
    samdb = None

    def setUp(self):
        super().setUp()
        self.samdb = self.getSamDB("-H", "ldap://%s" % os.environ["DC_SERVER"],
                                   "-U%s%%%s" % (os.environ["DC_USERNAME"], os.environ["DC_PASSWORD"]))
        self.domain_dn = self.samdb.domain_dn()

    def tearDown(self):
        super().tearDown()
        # Reset the values we might have changed.
        ds_dn = "CN=Directory Service,CN=Windows NT,CN=Services,CN=Configuration"
        m = ldb.Message()
        m.dn = ldb.Dn(self.samdb, "%s,%s" % (ds_dn, self.domain_dn))
        m['dsheuristics'] = ldb.MessageElement(
            '0000000', ldb.FLAG_MOD_REPLACE, 'dsheuristics')

        self.samdb.modify(m)

    def test_display(self):
        """Tests that we can display forest settings"""
        (result, out, err) = self.runcmd("forest",
                                         "directory_service",
                                         "show",
                                         "-H", "ldap://%s" % os.environ["DC_SERVER"],
                                         "-U%s%%%s" % (os.environ["DC_USERNAME"],
                                                       os.environ["DC_PASSWORD"]))

        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        self.assertIn("dsheuristics: <NO VALUE>", out)

    def test_modify_dsheuristics(self):
        """Test that we can modify the dsheuristics setting"""

        (result, out, err) = self.runcmd("forest",
                                         "directory_service",
                                         "dsheuristics",
                                         "0000002",
                                         "-H", "ldap://%s" % os.environ["DC_SERVER"],
                                         "-U%s%%%s" % (os.environ["DC_USERNAME"],
                                                       os.environ["DC_PASSWORD"]))

        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        self.assertIn("set dsheuristics: 0000002", out)



# Unix SMB/CIFS implementation.
# Copyright (C) Jelmer Vernooij <jelmer@samba.org> 2007
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for samba.upgrade."""

from samba.upgrade import import_wins
from samba.tests import LdbTestCase


class WinsUpgradeTests(LdbTestCase):

    def test_upgrade(self):
        winsdb = {
            "FOO#20": (200, ["127.0.0.1", "127.0.0.2"], 0x60)
        }
        import_wins(self.ldb, winsdb)

        self.assertEqual(
            ['name=FOO,type=0x20'],
            [str(m.dn) for m in
                self.ldb.search(expression="(objectClass=winsRecord)")])

    def test_version(self):
        import_wins(self.ldb, {})
        self.assertEqual("VERSION",
                          str(self.ldb.search(expression="(objectClass=winsMaxVersion)")[0]["cn"]))



# Unix SMB/CIFS implementation.
#
# Tests for samba-tool domain claim management
#
# Copyright (C) Catalyst.Net Ltd. 2023
#
# Written by Rob van der Linde <rob@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import json
import os

from ldb import SCOPE_ONELEVEL
from samba.sd_utils import SDUtils

from .base import SambaToolCmdTest

# List of claim value types we should expect to see.
VALUE_TYPES = [
    "Date Time",
    "Multi-valued Choice",
    "Multi-valued Text",
    "Number",
    "Ordered List",
    "Single-valued Choice",
    "Text",
    "Yes/No"
]

HOST = "ldap://{DC_SERVER}".format(**os.environ)
CREDS = "-U{DC_USERNAME}%{DC_PASSWORD}".format(**os.environ)


class BaseClaimCmdTest(SambaToolCmdTest):
    """Base class for claim types and claim value types tests."""

    @classmethod
    def setUpClass(cls):
        cls.samdb = cls.getSamDB("-H", HOST, CREDS)
        super().setUpClass()

    @classmethod
    def setUpTestData(cls):
        cls.create_claim_type("accountExpires", name="expires",
                              classes=["user"])
        cls.create_claim_type("department", name="dept", classes=["user"],
                              protect=True)
        cls.create_claim_type("carLicense", name="plate", classes=["user"],
                              disable=True)

    def get_services_dn(self):
        """Returns Services DN."""
        services_dn = self.samdb.get_config_basedn()
        services_dn.add_child("CN=Services")
        return services_dn

    def get_claim_types_dn(self):
        """Returns the Claim Types DN."""
        claim_types_dn = self.get_services_dn()
        claim_types_dn.add_child("CN=Claim Types,CN=Claims Configuration")
        return claim_types_dn

    @classmethod
    def _run(cls, *argv):
        """Override _run, so we don't always have to pass host and creds."""
        args = list(argv)
        args.extend(["-H", HOST, CREDS])
        return super()._run(*args)

    runcmd = _run
    runsubcmd = _run

    @classmethod
    def create_claim_type(cls, attribute, name=None, description=None,
                          classes=None, disable=False, protect=False):
        """Create a claim type using the samba-tool command."""

        # if name is specified it will override the attribute name
        display_name = name or attribute

        # base command for create claim-type
        cmd = ["domain", "claim", "claim-type",
               "create", "--attribute", attribute]

        # list of classes (applies_to)
        if classes is not None:
            cmd.extend([f"--class={name}" for name in classes])

        # optional attributes
        if name is not None:
            cmd.append(f"--name={name}")
        if description is not None:
            cmd.append(f"--description={description}")
        if disable:
            cmd.append("--disable")
        if protect:
            cmd.append("--protect")

        result, out, err = cls.runcmd(*cmd)
        assert result is None
        assert out.startswith("Created claim type")
        cls.addClassCleanup(cls.delete_claim_type, name=display_name, force=True)
        return display_name

    @classmethod
    def delete_claim_type(cls, name, force=False):
        """Delete claim type by display name."""
        cmd = ["domain", "claim", "claim-type", "delete", "--name", name]

        # Force-delete protected claim type.
        if force:
            cmd.append("--force")

        result, out, err = cls.runcmd(*cmd)
        assert result is None
        assert "Deleted claim type" in out

    def get_claim_type(self, name):
        """Get claim type by display name."""
        claim_types_dn = self.get_claim_types_dn()

        result = self.samdb.search(base=claim_types_dn,
                                   scope=SCOPE_ONELEVEL,
                                   expression=f"(displayName={name})")

        if len(result) == 1:
            return result[0]


class ClaimTypeCmdTestCase(BaseClaimCmdTest):
    """Tests for the claim-type command."""

    def test_list(self):
        """Test listing claim types in list format."""
        result, out, err = self.runcmd("domain", "claim", "claim-type", "list")
        self.assertIsNone(result, msg=err)

        expected_claim_types = ["expires", "dept", "plate"]

        for claim_type in expected_claim_types:
            self.assertIn(claim_type, out)

    def test_list__json(self):
        """Test listing claim types in JSON format."""
        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "list", "--json")
        self.assertIsNone(result, msg=err)

        # we should get valid json
        json_result = json.loads(out)
        claim_types = list(json_result.keys())

        expected_claim_types = ["expires", "dept", "plate"]

        for claim_type in expected_claim_types:
            self.assertIn(claim_type, claim_types)

    def test_view(self):
        """Test viewing a single claim type."""
        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "view", "--name", "expires")
        self.assertIsNone(result, msg=err)

        # we should get valid json
        claim_type = json.loads(out)

        # check a few fields only
        self.assertEqual(claim_type["displayName"], "expires")
        self.assertEqual(claim_type["description"], "Account-Expires")

    def test_view__name_missing(self):
        """Test view claim type without --name is handled."""
        result, out, err = self.runcmd("domain", "claim", "claim-type", "view")
        self.assertEqual(result, -1)
        self.assertIn("Argument --name is required.", err)

    def test_view__notfound(self):
        """Test viewing claim type that doesn't exist is handled."""
        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "view", "--name", "doesNotExist")
        self.assertEqual(result, -1)
        self.assertIn("Claim type doesNotExist not found.", err)

    def test_create(self):
        """Test creating several known attributes as claim types.

        The point is to test it against the various datatypes that could
        be found, but not include every known attribute.
        """
        # We just need to test a few different data types for attributes,
        # there is no need to test every known attribute.
        claim_types = [
            "adminCount",
            "accountExpires",
            "department",
            "carLicense",
            "msDS-PrimaryComputer",
            "isDeleted",
        ]

        # Each known attribute must be in the schema.
        for attribute in claim_types:
            # Use a different name, so we don't clash with existing attributes.
            name = "test_create_" + attribute

            self.addCleanup(self.delete_claim_type, name=name, force=True)

            result, out, err = self.runcmd("domain", "claim", "claim-type",
                                           "create",
                                           "--attribute", attribute,
                                           "--name", name,
                                           "--class=user")
            self.assertIsNone(result, msg=err)

            # It should have used the attribute name as displayName.
            claim_type = self.get_claim_type(name)
            self.assertEqual(str(claim_type["displayName"]), name)
            self.assertEqual(str(claim_type["Enabled"]), "TRUE")
            self.assertEqual(str(claim_type["objectClass"][-1]), "msDS-ClaimType")
            self.assertEqual(str(claim_type["msDS-ClaimSourceType"]), "AD")

    def test_create__boolean(self):
        """Test adding a known boolean attribute and check its type."""
        self.addCleanup(self.delete_claim_type, name="boolAttr", force=True)

        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "create", "--attribute=msNPAllowDialin",
                                       "--name=boolAttr", "--class=user")

        self.assertIsNone(result, msg=err)
        claim_type = self.get_claim_type("boolAttr")
        self.assertEqual(str(claim_type["displayName"]), "boolAttr")
        self.assertEqual(str(claim_type["msDS-ClaimValueType"]), "6")

    def test_create__number(self):
        """Test adding a known numeric attribute and check its type."""
        self.addCleanup(self.delete_claim_type, name="intAttr", force=True)

        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "create", "--attribute=adminCount",
                                       "--name=intAttr", "--class=user")

        self.assertIsNone(result, msg=err)
        claim_type = self.get_claim_type("intAttr")
        self.assertEqual(str(claim_type["displayName"]), "intAttr")
        self.assertEqual(str(claim_type["msDS-ClaimValueType"]), "1")

    def test_create__text(self):
        """Test adding a known text attribute and check its type."""
        self.addCleanup(self.delete_claim_type, name="textAttr", force=True)

        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "create", "--attribute=givenName",
                                       "--name=textAttr", "--class=user")

        self.assertIsNone(result, msg=err)
        claim_type = self.get_claim_type("textAttr")
        self.assertEqual(str(claim_type["displayName"]), "textAttr")
        self.assertEqual(str(claim_type["msDS-ClaimValueType"]), "3")

    def test_create__disabled(self):
        """Test adding a disabled attribute."""
        self.addCleanup(self.delete_claim_type, name="disabledAttr", force=True)

        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "create", "--attribute=msTSHomeDrive",
                                       "--name=disabledAttr", "--class=user",
                                       "--disable")

        self.assertIsNone(result, msg=err)
        claim_type = self.get_claim_type("disabledAttr")
        self.assertEqual(str(claim_type["displayName"]), "disabledAttr")
        self.assertEqual(str(claim_type["Enabled"]), "FALSE")

    def test_create__protected(self):
        """Test adding a protected attribute."""
        self.addCleanup(self.delete_claim_type, name="protectedAttr", force=True)

        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "create", "--attribute=mobile",
                                       "--name=protectedAttr", "--class=user",
                                       "--protect")

        self.assertIsNone(result, msg=err)
        claim_type = self.get_claim_type("protectedAttr")
        self.assertEqual(str(claim_type["displayName"]), "protectedAttr")

        # Check if the claim type is protected from accidental deletion.
        utils = SDUtils(self.samdb)
        desc = utils.get_sd_as_sddl(claim_type["dn"])
        self.assertIn("(D;;DTSD;;;WD)", desc)

    def test_create__classes(self):
        """Test adding an attribute applied to different classes."""
        schema_dn = self.samdb.get_schema_basedn()
        user_dn = f"CN=User,{schema_dn}"
        computer_dn = f"CN=Computer,{schema_dn}"

        # --class=user
        self.addCleanup(self.delete_claim_type, name="streetName", force=True)
        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "create", "--attribute=street",
                                       "--name=streetName", "--class=user")
        self.assertIsNone(result, msg=err)
        claim_type = self.get_claim_type("streetName")
        applies_to = [str(dn) for dn in claim_type["msDS-ClaimTypeAppliesToClass"]]
        self.assertEqual(str(claim_type["displayName"]), "streetName")
        self.assertEqual(len(applies_to), 1)
        self.assertIn(user_dn, applies_to)
        self.assertNotIn(computer_dn, applies_to)

        # --class=computer
        self.addCleanup(self.delete_claim_type, name="ext", force=True)
        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "create", "--attribute=extensionName",
                                       "--name=ext", "--class=computer")
        self.assertIsNone(result, msg=err)
        claim_type = self.get_claim_type("ext")
        applies_to = [str(dn) for dn in claim_type["msDS-ClaimTypeAppliesToClass"]]
        self.assertEqual(str(claim_type["displayName"]), "ext")
        self.assertEqual(len(applies_to), 1)
        self.assertNotIn(user_dn, applies_to)
        self.assertIn(computer_dn, applies_to)

        # --class=user --class=computer
        self.addCleanup(self.delete_claim_type,
                        name="primaryComputer", force=True)
        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "create", "--attribute=msDS-PrimaryComputer",
                                       "--name=primaryComputer", "--class=user",
                                       "--class=computer")
        self.assertIsNone(result, msg=err)
        claim_type = self.get_claim_type("primaryComputer")
        applies_to = [str(dn) for dn in claim_type["msDS-ClaimTypeAppliesToClass"]]
        self.assertEqual(str(claim_type["displayName"]), "primaryComputer")
        self.assertEqual(len(applies_to), 2)
        self.assertIn(user_dn, applies_to)
        self.assertIn(computer_dn, applies_to)

        # No classes should raise CommandError.
        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "create", "--attribute=wWWHomePage",
                                       "--name=homepage")
        self.assertEqual(result, -1)
        self.assertIn("Argument --class is required.", err)

    def test__delete(self):
        """Test deleting a claim type that is not protected."""
        # Create non-protected claim type.
        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "create", "--attribute=msDS-SiteName",
                                       "--name=siteName", "--class=computer")
        self.assertIsNone(result, msg=err)
        claim_type = self.get_claim_type("siteName")
        self.assertIsNotNone(claim_type)

        # Do the deletion.
        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "delete", "--name=siteName")
        self.assertIsNone(result, msg=err)

        # Claim type shouldn't exist anymore.
        claim_type = self.get_claim_type("siteName")
        self.assertIsNone(claim_type)

    def test_delete__protected(self):
        """Test deleting a protected claim type, with and without --force."""
        # Create protected claim type.
        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "create", "--attribute=postalCode",
                                       "--name=postcode", "--class=user",
                                       "--protect")
        self.assertIsNone(result, msg=err)
        claim_type = self.get_claim_type("postcode")
        self.assertIsNotNone(claim_type)

        # Do the deletion.
        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "delete", "--name=postcode")
        self.assertEqual(result, -1)

        # Claim type should still exist.
        claim_type = self.get_claim_type("postcode")
        self.assertIsNotNone(claim_type)

        # Try a force delete instead.
        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "delete", "--name=postcode", "--force")
        self.assertIsNone(result, msg=err)

        # Claim type shouldn't exist anymore.
        claim_type = self.get_claim_type("siteName")
        self.assertIsNone(claim_type)

    def test_delete__notfound(self):
        """Test deleting a claim type that doesn't exist."""
        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "delete", "--name", "doesNotExist")
        self.assertEqual(result, -1)
        self.assertIn("Claim type doesNotExist not found.", err)

    def test_modify__description(self):
        """Test modifying a claim type description."""
        self.addCleanup(self.delete_claim_type, name="company", force=True)
        self.create_claim_type("company", classes=["user"])

        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "modify", "--name", "company",
                                       "--description=NewDescription")
        self.assertIsNone(result, msg=err)

        # Verify fields were changed.
        claim_type = self.get_claim_type("company")
        self.assertEqual(str(claim_type["description"]), "NewDescription")

    def test_modify__classes(self):
        """Test modify claim type classes."""
        schema_dn = self.samdb.get_schema_basedn()
        user_dn = f"CN=User,{schema_dn}"
        computer_dn = f"CN=Computer,{schema_dn}"

        self.addCleanup(self.delete_claim_type, name="seeAlso", force=True)
        self.create_claim_type("seeAlso", classes=["user"])

        # First try removing all classes which shouldn't be allowed.
        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "modify", "--name", "seeAlso",
                                       "--class=")
        self.assertEqual(result, -1)
        self.assertIn("Class name is required.", err)

        # Try changing it to just --class=computer first.
        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "modify", "--name", "seeAlso",
                                       "--class=computer")
        self.assertIsNone(result, msg=err)
        claim_type = self.get_claim_type("seeAlso")
        applies_to = [str(dn) for dn in claim_type["msDS-ClaimTypeAppliesToClass"]]
        self.assertNotIn(user_dn, applies_to)
        self.assertIn(computer_dn, applies_to)

        # Now try changing it to --class=user again.
        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "modify", "--name", "seeAlso",
                                       "--class=user")
        self.assertIsNone(result, msg=err)
        claim_type = self.get_claim_type("seeAlso")
        applies_to = [str(dn) for dn in claim_type["msDS-ClaimTypeAppliesToClass"]]
        self.assertIn(user_dn, applies_to)
        self.assertNotIn(computer_dn, applies_to)

        # Why not both?
        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "modify", "--name", "seeAlso",
                                       "--class=user", "--class=computer")
        self.assertIsNone(result, msg=err)
        claim_type = self.get_claim_type("seeAlso")
        applies_to = [str(dn) for dn in claim_type["msDS-ClaimTypeAppliesToClass"]]
        self.assertIn(user_dn, applies_to)
        self.assertIn(computer_dn, applies_to)

    def test_modify__enable_disable(self):
        """Test modify disabling and enabling a claim type."""
        self.addCleanup(self.delete_claim_type, name="catalogs", force=True)
        self.create_claim_type("catalogs", classes=["user"])

        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "modify", "--name", "catalogs",
                                       "--disable")
        self.assertIsNone(result, msg=err)

        # Check that claim type was disabled.
        claim_type = self.get_claim_type("catalogs")
        self.assertEqual(str(claim_type["Enabled"]), "FALSE")

        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "modify", "--name", "catalogs",
                                       "--enable")
        self.assertIsNone(result, msg=err)

        # Check that claim type was enabled.
        claim_type = self.get_claim_type("catalogs")
        self.assertEqual(str(claim_type["Enabled"]), "TRUE")

    def test_modify__protect_unprotect(self):
        """Test modify un-protecting and protecting a claim type."""
        self.addCleanup(self.delete_claim_type, name="pager", force=True)
        self.create_claim_type("pager", classes=["user"])

        utils = SDUtils(self.samdb)
        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "modify", "--name", "pager",
                                       "--protect")
        self.assertIsNone(result, msg=err)

        # Check that claim type was protected.
        claim_type = self.get_claim_type("pager")
        desc = utils.get_sd_as_sddl(claim_type["dn"])
        self.assertIn("(D;;DTSD;;;WD)", desc)

        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "modify", "--name", "pager",
                                       "--unprotect")
        self.assertIsNone(result, msg=err)

        # Check that claim type was unprotected.
        claim_type = self.get_claim_type("pager")
        desc = utils.get_sd_as_sddl(claim_type["dn"])
        self.assertNotIn("(D;;DTSD;;;WD)", desc)

    def test_modify__enable_disable_together(self):
        """Test modify claim type doesn't allow both --enable and --disable."""
        self.addCleanup(self.delete_claim_type,
                        name="businessCategory", force=True)
        self.create_claim_type("businessCategory", classes=["user"])

        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "modify", "--name", "businessCategory",
                                       "--enable", "--disable")
        self.assertEqual(result, -1)
        self.assertIn("--enable and --disable cannot be used together.", err)

    def test_modify__protect_unprotect_together(self):
        """Test modify claim type using both --protect and --unprotect."""
        self.addCleanup(self.delete_claim_type,
                        name="businessCategory", force=True)
        self.create_claim_type("businessCategory", classes=["user"])

        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "modify", "--name", "businessCategory",
                                       "--protect", "--unprotect")
        self.assertEqual(result, -1)
        self.assertIn("--protect and --unprotect cannot be used together.", err)

    def test_modify__notfound(self):
        """Test modify a claim type that doesn't exist."""
        result, out, err = self.runcmd("domain", "claim", "claim-type",
                                       "modify", "--name", "doesNotExist",
                                       "--description=NewDescription")
        self.assertEqual(result, -1)
        self.assertIn("Claim type doesNotExist not found.", err)


class ValueTypeCmdTestCase(BaseClaimCmdTest):
    """Tests for the value-type command."""

    def test_list(self):
        """Test listing claim value types in list format."""
        result, out, err = self.runcmd("domain", "claim", "value-type", "list")
        self.assertIsNone(result, msg=err)

        # base list of value types is there
        for value_type in VALUE_TYPES:
            self.assertIn(value_type, out)

    def test_list__json(self):
        """Test listing claim value types in JSON format."""
        result, out, err = self.runcmd("domain", "claim", "value-type",
                                       "list", "--json")
        self.assertIsNone(result, msg=err)

        # we should get valid json
        json_result = json.loads(out)
        value_types = list(json_result.keys())

        # base list of value types is there
        for value_type in VALUE_TYPES:
            self.assertIn(value_type, value_types)

    def test_view(self):
        """Test viewing a single claim value type."""
        result, out, err = self.runcmd("domain", "claim", "value-type",
                                       "view", "--name", "Text")
        self.assertIsNone(result, msg=err)

        # we should get valid json
        value_type = json.loads(out)

        # check a few fields only
        self.assertEqual(value_type["name"], "MS-DS-Text")
        self.assertEqual(value_type["displayName"], "Text")
        self.assertEqual(value_type["msDS-ClaimValueType"], 3)

    def test_view__name_missing(self):
        """Test viewing a claim value type with missing --name is handled."""
        result, out, err = self.runcmd("domain", "claim", "value-type", "view")
        self.assertEqual(result, -1)
        self.assertIn("Argument --name is required.", err)

    def test_view__notfound(self):
        """Test viewing a claim value type that doesn't exist is handled."""
        result, out, err = self.runcmd("domain", "claim", "value-type",
                                       "view", "--name", "doesNotExist")
        self.assertEqual(result, -1)
        self.assertIn("Value type doesNotExist not found.", err)



# Unix SMB/CIFS implementation.
#
# encoders: JSONEncoder class for dealing with object fields.
#
# Copyright (C) Catalyst.Net Ltd. 2023
#
# Written by Rob van der Linde <rob@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import json
from datetime import datetime
from decimal import Decimal
from enum import Enum

from ldb import Dn, MessageElement, Result

from samba.dcerpc.security import descriptor


class JSONEncoder(json.JSONEncoder):
    """Custom JSON encoder class to help out with some data types.

    For example, the json module has no idea how to encode a Dn object to str.
    Another common object that is handled is Decimal types.

    In addition, any objects that have a __json__ method will get called.
    """

    def default(self, obj):
        if isinstance(obj, (Decimal, Dn, Exception, MessageElement)):
            return str(obj)
        if isinstance(obj, Result):
            return obj.msgs
        elif isinstance(obj, Enum):
            return str(obj.value)
        elif isinstance(obj, datetime):
            return obj.isoformat()
        elif isinstance(obj, descriptor):
            return obj.as_sddl()
        elif getattr(obj, "__json__", None) and callable(obj.__json__):
            return obj.__json__()
        return super().default(obj)



# Unix SMB/CIFS implementation.
# Copyright (C) Jelmer Vernooij <jelmer@samba.org> 2011
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for option parsing.

"""

import optparse
from samba.getopt import (
    AUTO_USE_KERBEROS,
    DONT_USE_KERBEROS,
    MUST_USE_KERBEROS,
    parse_kerberos_arg_legacy,
    parse_kerberos_arg,
)
import samba.tests


class KerberosOptionTests(samba.tests.TestCase):

    def test_legacy_parse_true(self):
        self.assertEqual(
            MUST_USE_KERBEROS, parse_kerberos_arg_legacy("yes", "--kerberos"))
        self.assertEqual(
            MUST_USE_KERBEROS, parse_kerberos_arg_legacy("true", "--kerberos"))
        self.assertEqual(
            MUST_USE_KERBEROS, parse_kerberos_arg_legacy("1", "--kerberos"))

    def test_legacy_parse_false(self):
        self.assertEqual(
            DONT_USE_KERBEROS, parse_kerberos_arg_legacy("no", "--kerberos"))
        self.assertEqual(
            DONT_USE_KERBEROS, parse_kerberos_arg_legacy("false", "--kerberos"))
        self.assertEqual(
            DONT_USE_KERBEROS, parse_kerberos_arg_legacy("0", "--kerberos"))

    def test_legacy_parse_auto(self):
        self.assertEqual(
            AUTO_USE_KERBEROS, parse_kerberos_arg_legacy("auto", "--kerberos"))

    def test_legacy_parse_invalid(self):
        self.assertRaises(optparse.OptionValueError,
                          parse_kerberos_arg_legacy, "blah?", "--kerberos")

    def test_parse_valid(self):
        self.assertEqual(
            MUST_USE_KERBEROS, parse_kerberos_arg("required", "--use-kerberos"))
        self.assertEqual(
            AUTO_USE_KERBEROS, parse_kerberos_arg("desired", "--use-kerberos"))
        self.assertEqual(
            DONT_USE_KERBEROS, parse_kerberos_arg("off", "--use-kerberos"))

    def test_parse_invalid(self):
        self.assertRaises(optparse.OptionValueError,
                          parse_kerberos_arg, "wurst", "--use-kerberos")



#
# Unix SMB/CIFS implementation.
# Copyright (c) 2020      Andreas Schneider <asn@samba.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for samba.dcerpc., credentials and binding strings"""

import samba.tests
from samba.tests import RpcInterfaceTestCase, TestCase
from samba.dcerpc import lsa
import samba.dcerpc.security as security
from samba.credentials import Credentials, SMB_ENCRYPTION_REQUIRED, SMB_ENCRYPTION_OFF
from samba import NTSTATUSError

class RpcBindingTests(RpcInterfaceTestCase):

    def get_user_creds(self):
        c = Credentials()
        c.guess()
        domain = samba.tests.env_get_var_value('DOMAIN')
        username = samba.tests.env_get_var_value('USERNAME')
        password = samba.tests.env_get_var_value('PASSWORD')
        c.set_domain(domain)
        c.set_username(username)
        c.set_password(password)
        return c

    def test_smb3_dcerpc_no_encryption(self):
        creds = self.get_user_creds()
        creds.set_smb_encryption(SMB_ENCRYPTION_OFF)

        lp = self.get_loadparm()
        lp.set('client ipc max protocol', 'SMB3')
        lp.set('client ipc min protocol', 'SMB3')

        binding_string = ("ncacn_np:%s" % (samba.tests.env_get_var_value('SERVER')))
        lsa_conn = lsa.lsarpc(binding_string, lp, creds)
        self.assertFalse(lsa_conn.transport_encrypted())

        objectAttr = lsa.ObjectAttribute()
        objectAttr.sec_qos = lsa.QosInfo()

        pol_handle = lsa_conn.OpenPolicy2('',
                                          objectAttr,
                                          security.SEC_FLAG_MAXIMUM_ALLOWED)
        self.assertIsNotNone(pol_handle)

    def test_smb3_dcerpc_encryption(self):
        creds = self.get_user_creds()
        creds.set_smb_encryption(SMB_ENCRYPTION_REQUIRED)

        lp = self.get_loadparm()
        lp.set('client ipc max protocol', 'SMB3')
        lp.set('client ipc min protocol', 'SMB3')

        binding_string = ("ncacn_np:%s" % (samba.tests.env_get_var_value('SERVER')))
        lsa_conn = lsa.lsarpc(binding_string, lp, creds)
        self.assertTrue(lsa_conn.transport_encrypted())

        objectAttr = lsa.ObjectAttribute()
        objectAttr.sec_qos = lsa.QosInfo()

        pol_handle = lsa_conn.OpenPolicy2('',
                                          objectAttr,
                                          security.SEC_FLAG_MAXIMUM_ALLOWED)
        self.assertIsNotNone(pol_handle)

    def test_smb2_dcerpc_encryption(self):
        creds = self.get_user_creds()
        creds.set_smb_encryption(SMB_ENCRYPTION_REQUIRED)

        lp = self.get_loadparm()
        lp.set('client ipc max protocol', 'SMB2')
        lp.set('client ipc min protocol', 'SMB2')

        binding_string = ("ncacn_np:%s" % (samba.tests.env_get_var_value('SERVER')))
        self.assertRaises(NTSTATUSError, lsa.lsarpc, binding_string, lp, creds)

    def test_smb1_dcerpc_encryption(self):
        creds = self.get_user_creds()
        creds.set_smb_encryption(SMB_ENCRYPTION_REQUIRED)

        lp = self.get_loadparm()
        lp.set('client ipc max protocol', 'NT1')
        lp.set('client ipc min protocol', 'NT1')

        binding_string = ("ncacn_np:%s" % (samba.tests.env_get_var_value('SERVER')))
        self.assertRaises(NTSTATUSError, lsa.lsarpc, binding_string, lp, creds)



# -*- coding: utf-8 -*-

# Unix SMB/CIFS implementation.
# Copyright © Jelmer Vernooij <jelmer@samba.org> 2008
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


"""Network Data Representation (NDR) marshalling and unmarshalling."""


def ndr_pack(object) -> bytes:
    """Pack a NDR object.

    :param object: Object to pack
    :return: bytes object with marshalled object.
    """
    ndr_pack = getattr(object, "__ndr_pack__", None)
    if ndr_pack is None:
        raise TypeError("%r is not a NDR object" % object)
    return ndr_pack()


def ndr_unpack(cls, data:bytes, allow_remaining=False):
    """NDR unpack an object.

    :param cls: Class of the object to unpack
    :param data: Buffer to unpack
    :param allow_remaining: allows remaining data at the end (default=False)
    :return: Unpacked object
    """
    object = cls()
    ndr_unpack = getattr(object, "__ndr_unpack__", None)
    if ndr_unpack is None:
        raise TypeError("%r is not a NDR object" % object)
    ndr_unpack(data, allow_remaining=allow_remaining)
    return object


def ndr_print(object, print_secrets=False):
    ndr_print = getattr(object, "__ndr_print__", None)
    if ndr_print is None:
        raise TypeError(f"{object} is not a NDR object")
    return ndr_print(print_secrets=print_secrets)


def ndr_deepcopy(object):
    """Create a deep copy of a NDR object, using pack/unpack

    :param object: Object to copy
    :return: The object copy
    """
    ndr_pack = getattr(object, "__ndr_pack__", None)
    if ndr_pack is None:
        raise TypeError("%r is not a NDR object" % object)
    data = ndr_pack()
    cls = type(object)
    copy = cls()
    ndr_unpack = getattr(copy, "__ndr_unpack__", None)
    if ndr_unpack is None:
        raise TypeError("%r is not a NDR object" % copy)
    ndr_unpack(data, allow_remaining=False)
    return copy


def ndr_pack_in(object, bigendian=False, ndr64=False) -> bytes:
    """Pack the input of an NDR function object.

    :param object: Object to pack
    :param bigendian: use LIBNDR_FLAG_BIGENDIAN (default=False)
    :param ndr64: use LIBNDR_FLAG_NDR64 (default=False)
    :return: bytes object with marshalled object.
    """
    ndr_pack_in_fn = getattr(object, "__ndr_pack_in__", None)
    if ndr_pack_in_fn is None:
        raise TypeError("%r is not a NDR function object" % object)
    return ndr_pack_in_fn(bigendian=bigendian, ndr64=ndr64)


def ndr_unpack_in(object,
                  data: bytes,
                  bigendian=False,
                  ndr64=False,
                  allow_remaining=False):
    """Unpack the input of an NDR function object.

    :param cls: Class of the object to unpack
    :param data: Buffer to unpack
    :param bigendian: use LIBNDR_FLAG_BIGENDIAN (default=False)
    :param ndr64: use LIBNDR_FLAG_NDR64 (default=False)
    :param allow_remaining: allows remaining data at the end (default=False)
    :return: Unpacked object
    """
    ndr_unpack_in_fn = getattr(object, "__ndr_unpack_in__", None)
    if ndr_unpack_in_fn is None:
        raise TypeError("%r is not a NDR function object" % object)
    ndr_unpack_in_fn(data, bigendian=bigendian, ndr64=ndr64,
                     allow_remaining=allow_remaining)
    return object


def ndr_print_in(object, print_secrets=False):
    ndr_print_in_fn = getattr(object, "__ndr_print_in__", None)
    if ndr_print_in_fn is None:
        raise TypeError("%r is not a NDR function object" % object)
    return ndr_print_in_fn(print_secrets=print_secrets)


def ndr_pack_out(object, bigendian=False, ndr64=False):
    """Pack the output of an NDR function object.

    :param object: Object to pack
    :param bigendian: use LIBNDR_FLAG_BIGENDIAN (default=False)
    :param ndr64: use LIBNDR_FLAG_NDR64 (default=False)
    :return: String object with marshalled object.
    """
    ndr_pack_out_fn = getattr(object, "__ndr_pack_out__", None)
    if ndr_pack_out_fn is None:
        raise TypeError("%r is not a NDR function object" % object)
    return ndr_pack_out_fn(bigendian=bigendian, ndr64=ndr64)


def ndr_unpack_out(object, data, bigendian=False, ndr64=False, allow_remaining=False):
    """Unpack the output of an NDR function object.

    :param cls: Class of the object to unpack
    :param data: Buffer to unpack
    :param bigendian: use LIBNDR_FLAG_BIGENDIAN (default=False)
    :param ndr64: use LIBNDR_FLAG_NDR64 (default=False)
    :param allow_remaining: allows remaining data at the end (default=False)
    :return: Unpacked object
    """
    ndr_unpack_out_fn = getattr(object, "__ndr_unpack_out__", None)
    if ndr_unpack_out_fn is None:
        raise TypeError("%r is not a NDR function object" % object)
    ndr_unpack_out_fn(data, bigendian=bigendian, ndr64=ndr64,
                      allow_remaining=allow_remaining)
    return object


def ndr_print_out(object, print_secrets=False):
    ndr_print_out_fn = getattr(object, "__ndr_print_out__", None)
    if ndr_print_out_fn is None:
        raise TypeError("%r is not a NDR function object" % object)
    return ndr_print_out_fn(print_secrets=print_secrets)



# Unix SMB/CIFS implementation.
# Copyright (C) Andrew Bartlett <abartlet@samba.org> 2017
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for the Auth and AuthZ logging.
"""

import samba.tests
from samba.messaging import Messaging
from samba.dcerpc.messaging import MSG_AUTH_LOG, AUTH_EVENT_NAME
from samba.param import LoadParm
from samba import string_is_guid
import time
import json
import os


def default_msg_filter(msg):
    # When our authentication logging tests were written, these were the only
    # supported message types. The tests were built on the assumption that no
    # new types would be added, and violating this assumption will result in
    # many tests failing as they receive messages that they weren’t
    # expecting. To allow these tests to continue to pass, this default filter
    # makes sure that only messages for which the tests are prepared pass
    # though.
    default_supported_types = {
        "Authentication",
        "Authorization",
    }

    return msg['type'] in default_supported_types


class NoMessageException(Exception):
    pass


class AuthLogTestBase(samba.tests.TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        # connect to the server's messaging bus (we need to explicitly load a
        # different smb.conf here, because in all other respects this test
        # wants to act as a separate remote client)
        server_conf = os.getenv('SERVERCONFFILE')
        if server_conf:
            lp_ctx = LoadParm(filename_for_non_global_lp=server_conf)
        else:
            lp_ctx = samba.tests.env_loadparm()
        cls.msg_ctx = Messaging((1,), lp_ctx=lp_ctx)
        cls.msg_ctx.irpc_add_name(AUTH_EVENT_NAME)

        # Now switch back to using the client-side smb.conf. The tests will
        # use the first interface in the client.conf (we need to strip off
        # the subnet mask portion)
        lp_ctx = samba.tests.env_loadparm()
        client_ip_and_mask = lp_ctx.get('interfaces')[0]
        client_ip = client_ip_and_mask.split('/')[0]

        # the messaging ctx is the server's view of the world, so our own
        # client IP will be the remoteAddress when connections are logged
        cls.remoteAddress = client_ip

        def messageHandler(context, msgType, src, message):
            # This does not look like sub unit output and it
            # makes these tests much easier to debug.
            print(message)
            jsonMsg = json.loads(message)
            context["messages"].append(jsonMsg)

        cls.context = {"messages": []}
        cls.msg_handler_and_context = (messageHandler, cls.context)
        cls.msg_ctx.register(cls.msg_handler_and_context,
                             msg_type=MSG_AUTH_LOG)

        cls.server = os.environ["SERVER"]
        cls.connection = None

    @classmethod
    def tearDownClass(cls):
        cls.msg_ctx.deregister(cls.msg_handler_and_context,
                               msg_type=MSG_AUTH_LOG)
        cls.msg_ctx.irpc_remove_name(AUTH_EVENT_NAME)

        super().tearDownClass()

    def setUp(self):
        super().setUp()
        type(self).discardMessages()

    def isRemote(self, message):
        if self.remoteAddress is None:
            return True

        supported_types = {
            "Authentication",
            "Authorization",
            "KDC Authorization",
        }
        message_type = message["type"]
        if message_type in supported_types:
            remote = message[message_type]["remoteAddress"]
        else:
            return False

        try:
            addr = remote.split(":")
            return addr[1] == self.remoteAddress
        except IndexError:
            return False

    def waitForMessages(self, isLastExpectedMessage, connection=None, *,
                        msgFilter=default_msg_filter):
        """Wait for all the expected messages to arrive
        The connection is passed through to keep the connection alive
        until all the logging messages have been received.

        By default, only Authentication and Authorization messages will be
        returned, so that old tests continue to pass. To receive all messages,
        pass msgFilter=None.

        """

        messages = []
        while True:
            try:
                msg = self.nextMessage(msgFilter=msgFilter)
            except NoMessageException:
                return []

            messages.append(msg)
            if isLastExpectedMessage(msg):
                return messages

    def nextMessage(self, msgFilter=None):
        """Return the next relevant message, or throw a NoMessageException."""
        def is_relevant(msg):
            if not self.isRemote(msg):
                return False

            if msgFilter is None:
                return True

            return msgFilter(msg)

        messages = self.context['messages']

        while True:
            timeout = 2
            until = time.time() + timeout

            while not messages:
                # Fetch a new message from the messaging bus.

                current = time.time()
                if until < current:
                    break

                self.msg_ctx.loop_once(until - current)

            if not messages:
                raise NoMessageException('timed out looking for a message')

            # Grab the next message from the queue.
            msg = messages.pop(0)
            if is_relevant(msg):
                return msg

    # Discard any previously queued messages.
    @classmethod
    def discardMessages(cls):
        messages = cls.context["messages"]

        while True:
            messages.clear()

            # tevent presumably has other tasks to run, so we might need two or
            # three loops before a message comes through.
            for _ in range(5):
                cls.msg_ctx.loop_once(0.001)

            if not messages:
                # No new messages. We’ve probably got them all.
                break

    # Remove any NETLOGON authentication messages
    # NETLOGON is only performed once per session, so to avoid ordering
    # dependencies within the tests it's best to strip out NETLOGON messages.
    #
    def remove_netlogon_messages(self, messages):
        def is_not_netlogon(msg):
            if "Authentication" not in msg:
                return True
            sd = msg["Authentication"]["serviceDescription"]
            return sd != "NETLOGON"

        return list(filter(is_not_netlogon, messages))

    def is_guid(self, guid):
        """Is the supplied GUID string correctly formatted"""
        return string_is_guid(guid)



# -*- coding: utf-8 -*-
#
# Unix SMB/CIFS implementation.
# Copyright © Andrew Bartlett <abartlet@samba.org> 2021
# Copyright (C) Catalyst IT Ltd. 2017
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for samba.dcerpc.lsa."""

from samba.dcerpc import lsa, security
from samba.credentials import Credentials
from samba.tests import TestCase
from samba.dcerpc.security import dom_sid
from samba import NTSTATUSError
from samba.ntstatus import (
    NT_STATUS_OK,
    NT_STATUS_ACCESS_DENIED,
    NT_STATUS_NONE_MAPPED,
    NT_STATUS_SOME_NOT_MAPPED,
    NT_STATUS_INVALID_SID,
)
import samba.tests

class LsaTests(TestCase):

    def setUp(self):
        self.lp = self.get_loadparm()
        self.server = samba.tests.env_get_var_value('SERVER')

    def test_lsa_LookupSids3_multiple(self):
        machine_creds = Credentials()
        machine_creds.guess(self.lp)
        machine_creds.set_machine_account()

        c = lsa.lsarpc(
            "ncacn_ip_tcp:%s[schannel,seal]" % self.server,
            self.lp,
            machine_creds)

        sids = lsa.SidArray()
        sid = lsa.SidPtr()
        # Need a set
        x = dom_sid("S-1-5-7")
        sid.sid = x
        sids.sids = [sid]
        sids.num_sids = 1
        names = lsa.TransNameArray2()
        level = lsa.LSA_LOOKUP_NAMES_ALL
        count = 0
        lookup_options = lsa.LSA_LOOKUP_OPTION_SEARCH_ISOLATED_NAMES
        client_revision = lsa.LSA_CLIENT_REVISION_2

        # We want to run LookupSids3 multiple times on the same
        # connection as we have code to re-use the sam.ldb and we need
        # to check things work for the second request.
        (domains, names, count) = c.LookupSids3(sids, names, level, count, lookup_options, client_revision)
        self.assertEqual(count, 1)
        self.assertEqual(names.count, 1)
        self.assertEqual(names.names[0].name.string,
                         "ANONYMOUS LOGON")
        (domains2, names2, count2) = c.LookupSids3(sids, names, level, count, lookup_options, client_revision)
        self.assertEqual(count2, 1)
        self.assertEqual(names2.count, 1)
        self.assertEqual(names2.names[0].name.string,
                         "ANONYMOUS LOGON")

        # Just looking for any exceptions in the last couple of loops
        c.LookupSids3(sids, names, level, count, lookup_options, client_revision)
        c.LookupSids3(sids, names, level, count, lookup_options, client_revision)

    def test_lsa_LookupSids3_multiple_conns(self):
        machine_creds = Credentials()
        machine_creds.guess(self.lp)
        machine_creds.set_machine_account()

        c = lsa.lsarpc(
            "ncacn_ip_tcp:%s[schannel,seal]" % self.server,
            self.lp,
            machine_creds)

        sids = lsa.SidArray()
        sid = lsa.SidPtr()
        # Need a set
        x = dom_sid("S-1-5-7")
        sid.sid = x
        sids.sids = [sid]
        sids.num_sids = 1
        names = lsa.TransNameArray2()
        level = lsa.LSA_LOOKUP_NAMES_ALL
        count = 0
        lookup_options = lsa.LSA_LOOKUP_OPTION_SEARCH_ISOLATED_NAMES
        client_revision = lsa.LSA_CLIENT_REVISION_2

        # We want to run LookupSids3, and then again on a new
        # connection to show that we don't have an issue with the DB
        # being tied to the wrong connection.
        (domains, names, count) = c.LookupSids3(sids,
                                                names,
                                                level,
                                                count,
                                                lookup_options,
                                                client_revision)
        self.assertEqual(count, 1)
        self.assertEqual(names.count, 1)
        self.assertEqual(names.names[0].name.string,
                         "ANONYMOUS LOGON")

        c = lsa.lsarpc(
            "ncacn_ip_tcp:%s[schannel,seal]" % self.server,
            self.lp,
            machine_creds)

        (domains, names, count) = c.LookupSids3(sids,
                                                names,
                                                level,
                                                count,
                                                lookup_options,
                                                client_revision)
        self.assertEqual(count, 1)
        self.assertEqual(names.count, 1)
        self.assertEqual(names.names[0].name.string,
                         "ANONYMOUS LOGON")


    def test_lsa_LookupNames4_LookupSids3_multiple(self):
        """
        Test by going back and forward between real DB lookups
        name->sid->name to ensure the sam.ldb handle is fine once
        shared
        """

        machine_creds = Credentials()
        machine_creds.guess(self.lp)
        machine_creds.set_machine_account()

        c_normal = lsa.lsarpc(
            "ncacn_np:%s[seal]" % self.server,
            self.lp,
            machine_creds)

        username, domain = c_normal.GetUserName(None, None, None)

        c = lsa.lsarpc(
            "ncacn_ip_tcp:%s[schannel,seal]" % self.server,
            self.lp,
            machine_creds)

        sids  = lsa.TransSidArray3()
        names = [username]
        level = lsa.LSA_LOOKUP_NAMES_ALL
        count = 0
        lookup_options = lsa.LSA_LOOKUP_OPTION_SEARCH_ISOLATED_NAMES
        client_revision = lsa.LSA_CLIENT_REVISION_2
        (domains, sids, count) = c.LookupNames4(names,
                                                sids,
                                                level,
                                                count,
                                                lookup_options,
                                                client_revision)

        # Another lookup on the same connection, will re-used the
        # server-side implicit state handle on the connection
        (domains, sids, count) = c.LookupNames4(names,
                                                sids,
                                                level,
                                                count,
                                                lookup_options,
                                                client_revision)

        self.assertEqual(count, 1)
        self.assertEqual(sids.count, 1)

        # Now look the SIDs back up
        names = lsa.TransNameArray2()
        sid = lsa.SidPtr()
        sid.sid = sids.sids[0].sid
        lookup_sids = lsa.SidArray()
        lookup_sids.sids = [sid]
        lookup_sids.num_sids = 1
        level = lsa.LSA_LOOKUP_NAMES_ALL
        count = 1
        lookup_options = 0
        client_revision = lsa.LSA_CLIENT_REVISION_2

        (domains, names, count) = c.LookupSids3(lookup_sids,
                                                names,
                                                level,
                                                count,
                                                lookup_options,
                                                client_revision)
        self.assertEqual(count, 1)
        self.assertEqual(names.count, 1)
        self.assertEqual(names.names[0].name.string,
                         username.string)

        # And once more just to be sure, just checking for a fault
        sids  = lsa.TransSidArray3()
        names = [username]
        level = lsa.LSA_LOOKUP_NAMES_ALL
        count = 0
        lookup_options = lsa.LSA_LOOKUP_OPTION_SEARCH_ISOLATED_NAMES
        client_revision = lsa.LSA_CLIENT_REVISION_2
        (domains, sids, count) = c.LookupNames4(names,
                                                sids,
                                                level,
                                                count,
                                                lookup_options,
                                                client_revision)


    def test_lsa_LookupNames4_multiple_conns(self):
        """
        Test by going back and forward between real DB lookups
        name->sid->name to ensure the sam.ldb handle is fine once
        shared
        """

        machine_creds = Credentials()
        machine_creds.guess(self.lp)
        machine_creds.set_machine_account()

        c_normal = lsa.lsarpc(
            "ncacn_np:%s[seal]" % self.server,
            self.lp,
            machine_creds)

        username, domain = c_normal.GetUserName(None, None, None)

        c = lsa.lsarpc(
            "ncacn_ip_tcp:%s[schannel,seal]" % self.server,
            self.lp,
            machine_creds)

        sids  = lsa.TransSidArray3()
        names = [username]
        level = lsa.LSA_LOOKUP_NAMES_ALL
        count = 0
        lookup_options = lsa.LSA_LOOKUP_OPTION_SEARCH_ISOLATED_NAMES
        client_revision = lsa.LSA_CLIENT_REVISION_2
        (domains, sids, count) = c.LookupNames4(names,
                                                sids,
                                                level,
                                                count,
                                                lookup_options,
                                                client_revision)

        c = lsa.lsarpc(
            "ncacn_ip_tcp:%s[schannel,seal]" % self.server,
            self.lp,
            machine_creds)

        sids  = lsa.TransSidArray3()
        names = [username]
        level = lsa.LSA_LOOKUP_NAMES_ALL
        count = 0
        lookup_options = lsa.LSA_LOOKUP_OPTION_SEARCH_ISOLATED_NAMES
        client_revision = lsa.LSA_CLIENT_REVISION_2
        (domains, sids, count) = c.LookupNames4(names,
                                                sids,
                                                level,
                                                count,
                                                lookup_options,
                                                client_revision)

    def test_lsa_LookupNames4_without_schannel(self):

        machine_creds = Credentials()
        machine_creds.guess(self.lp)
        machine_creds.set_machine_account()

        c_normal = lsa.lsarpc(
            "ncacn_np:%s[seal]" % self.server,
            self.lp,
            machine_creds)

        username, domain = c_normal.GetUserName(None, None, None)

        sids  = lsa.TransSidArray3()
        names = [username]
        level = lsa.LSA_LOOKUP_NAMES_ALL
        count = 0
        lookup_options = lsa.LSA_LOOKUP_OPTION_SEARCH_ISOLATED_NAMES
        client_revision = lsa.LSA_CLIENT_REVISION_2

        with self.assertRaises(NTSTATUSError) as e:
            c_normal.LookupNames4(names,
                                  sids,
                                  level,
                                  count,
                                  lookup_options,
                                  client_revision)
        if (e.exception.args[0] != NT_STATUS_ACCESS_DENIED):
            raise AssertionError("LookupNames4 without schannel must fail with ACCESS_DENIED")

    def test_lsa_LookupSids3_without_schannel(self):
        machine_creds = Credentials()
        machine_creds.guess(self.lp)
        machine_creds.set_machine_account()

        c = lsa.lsarpc(
            "ncacn_ip_tcp:%s[seal]" % self.server,
            self.lp,
            machine_creds)

        sids = lsa.SidArray()
        sid = lsa.SidPtr()
        # Need a set
        x = dom_sid("S-1-5-7")
        sid.sid = x
        sids.sids = [sid]
        sids.num_sids = 1
        names = lsa.TransNameArray2()
        level = lsa.LSA_LOOKUP_NAMES_ALL
        count = 0
        lookup_options = lsa.LSA_LOOKUP_OPTION_SEARCH_ISOLATED_NAMES
        client_revision = lsa.LSA_CLIENT_REVISION_2

        with self.assertRaises(NTSTATUSError) as e:
            c.LookupSids3(sids,
                          names,
                          level,
                          count,
                          lookup_options,
                          client_revision)
        if (e.exception.args[0] != NT_STATUS_ACCESS_DENIED):
            raise AssertionError("LookupSids3 without schannel must fail with ACCESS_DENIED")

    def test_lsa_LookupSids2_none_mapped(self):
        machine_creds = Credentials()
        machine_creds.guess(self.lp)
        machine_creds.set_machine_account()

        c = lsa.lsarpc(
            "ncacn_np:%s[print]" % self.server,
            self.lp,
            machine_creds,
            raise_result_exceptions=False)

        objectAttr = lsa.ObjectAttribute()
        objectAttr.sec_qos = lsa.QosInfo()

        (pol_handle, status) = c.OpenPolicy2('',
                                             objectAttr,
                                             security.SEC_FLAG_MAXIMUM_ALLOWED)
        self.assertEqual(status[0], NT_STATUS_OK)
        self.assertIsNotNone(pol_handle)

        x0 = dom_sid("S-1-3-66")
        sid0 = lsa.SidPtr()
        sid0.sid = x0
        x1 = dom_sid("S-1-3-77")
        sid1 = lsa.SidPtr()
        sid1.sid = x1
        x2 = dom_sid("S-1-3-88")
        sid2 = lsa.SidPtr()
        sid2.sid = x2
        x3 = dom_sid("S-1-3-99")
        sid3 = lsa.SidPtr()
        sid3.sid = x3
        sids = lsa.SidArray()
        sids.sids = [sid0,sid1,sid2,sid3]
        sids.num_sids = 4

        names = lsa.TransNameArray2()
        level = lsa.LSA_LOOKUP_NAMES_ALL
        count = 0
        lookup_options = lsa.LSA_LOOKUP_OPTION_SEARCH_ISOLATED_NAMES
        client_revision = lsa.LSA_CLIENT_REVISION_2

        (domains, names, count, status) = c.LookupSids2(pol_handle,
                                                        sids,
                                                        names,
                                                        level,
                                                        count,
                                                        lookup_options,
                                                        client_revision)
        self.assertEqual(status[0], NT_STATUS_NONE_MAPPED)
        self.assertEqual(count, 0)
        self.assertIsNotNone(domains)
        self.assertEqual(domains.count, 0)
        self.assertIsNotNone(names)
        self.assertEqual(names.count, 4)
        self.assertEqual(names.names[0].sid_type, lsa.SID_NAME_UNKNOWN)
        self.assertEqual(names.names[0].name.string, str(x0))
        self.assertEqual(names.names[0].sid_index, 0xffffffff)
        self.assertEqual(names.names[1].sid_type, lsa.SID_NAME_UNKNOWN)
        self.assertEqual(names.names[1].name.string, str(x1))
        self.assertEqual(names.names[1].sid_index, 0xffffffff)
        self.assertEqual(names.names[2].sid_type, lsa.SID_NAME_UNKNOWN)
        self.assertEqual(names.names[2].name.string, str(x2))
        self.assertEqual(names.names[2].sid_index, 0xffffffff)
        self.assertEqual(names.names[3].sid_type, lsa.SID_NAME_UNKNOWN)
        self.assertEqual(names.names[3].name.string, str(x3))
        self.assertEqual(names.names[3].sid_index, 0xffffffff)

        return

    def test_lsa_LookupSids2_some_not_mapped(self):
        machine_creds = Credentials()
        machine_creds.guess(self.lp)
        machine_creds.set_machine_account()

        c = lsa.lsarpc(
            "ncacn_np:%s[print]" % self.server,
            self.lp,
            machine_creds,
            raise_result_exceptions=False)

        objectAttr = lsa.ObjectAttribute()
        objectAttr.sec_qos = lsa.QosInfo()

        (pol_handle, status) = c.OpenPolicy2('',
                                             objectAttr,
                                             security.SEC_FLAG_MAXIMUM_ALLOWED)
        self.assertEqual(status[0], NT_STATUS_OK)
        self.assertIsNotNone(pol_handle)

        dx0 = dom_sid("S-1-3")

        x0 = dom_sid("S-1-3-66")
        sid0 = lsa.SidPtr()
        sid0.sid = x0
        x1 = dom_sid("S-1-3-0")
        sid1 = lsa.SidPtr()
        sid1.sid = x1
        x2 = dom_sid("S-1-3")
        sid2 = lsa.SidPtr()
        sid2.sid = x2
        x3 = dom_sid("S-1-3-99")
        sid3 = lsa.SidPtr()
        sid3.sid = x3
        sids = lsa.SidArray()
        sids.sids = [sid0,sid1,sid2,sid3]
        sids.num_sids = 4

        names = lsa.TransNameArray2()
        level = lsa.LSA_LOOKUP_NAMES_ALL
        count = 0
        lookup_options = lsa.LSA_LOOKUP_OPTION_SEARCH_ISOLATED_NAMES
        client_revision = lsa.LSA_CLIENT_REVISION_2

        (domains, names, count, status) = c.LookupSids2(pol_handle,
                                                        sids,
                                                        names,
                                                        level,
                                                        count,
                                                        lookup_options,
                                                        client_revision)
        self.assertEqual(status[0], NT_STATUS_SOME_NOT_MAPPED)
        self.assertEqual(count, 1)
        self.assertIsNotNone(domains)
        self.assertEqual(domains.count, 1)
        self.assertEqual(domains.domains[0].name.string, "")
        self.assertEqual(domains.domains[0].sid, dx0)
        self.assertIsNotNone(names)
        self.assertEqual(names.count, 4)
        self.assertEqual(names.names[0].sid_type, lsa.SID_NAME_UNKNOWN)
        self.assertEqual(names.names[0].name.string, str(x0))
        self.assertEqual(names.names[0].sid_index, 0xffffffff)
        self.assertEqual(names.names[1].sid_type, lsa.SID_NAME_WKN_GRP)
        self.assertEqual(names.names[1].name.string, "CREATOR OWNER")
        self.assertEqual(names.names[1].sid_index, 0)
        self.assertEqual(names.names[2].sid_type, lsa.SID_NAME_UNKNOWN)
        self.assertEqual(names.names[2].name.string, str(x2))
        self.assertEqual(names.names[2].sid_index, 0xffffffff)
        self.assertEqual(names.names[3].sid_type, lsa.SID_NAME_UNKNOWN)
        self.assertEqual(names.names[3].name.string, str(x3))
        self.assertEqual(names.names[3].sid_index, 0xffffffff)

        return

    def test_lsa_LookupSids2_invalid_sid(self):
        machine_creds = Credentials()
        machine_creds.guess(self.lp)
        machine_creds.set_machine_account()

        c = lsa.lsarpc(
            "ncacn_np:%s[print]" % self.server,
            self.lp,
            machine_creds,
            raise_result_exceptions=False)

        objectAttr = lsa.ObjectAttribute()
        objectAttr.sec_qos = lsa.QosInfo()

        (pol_handle, status) = c.OpenPolicy2('',
                                             objectAttr,
                                             security.SEC_FLAG_MAXIMUM_ALLOWED)
        self.assertEqual(status[0], NT_STATUS_OK)
        self.assertIsNotNone(pol_handle)

        dx0 = dom_sid("S-1-3")

        x0 = dom_sid("S-1-3-66")
        sid0 = lsa.SidPtr()
        sid0.sid = x0
        x1 = dom_sid("S-1-3-77")
        sid1 = lsa.SidPtr()
        sid1.sid = x1
        x2 = dom_sid("S-1-3")
        sid2 = lsa.SidPtr()
        sid2.sid = x2
        x3 = dom_sid("S-1-3-99")
        sid3 = lsa.SidPtr()
        sid3.sid = x3
        sids = lsa.SidArray()
        sids.sids = [sid0,sid1,sid2,sid3]
        sids.num_sids = 4

        names = lsa.TransNameArray2()
        level = lsa.LSA_LOOKUP_NAMES_ALL
        count = 0
        lookup_options = lsa.LSA_LOOKUP_OPTION_SEARCH_ISOLATED_NAMES
        client_revision = lsa.LSA_CLIENT_REVISION_2

        (domains, names, count, status) = c.LookupSids2(pol_handle,
                                                        sids,
                                                        names,
                                                        level,
                                                        count,
                                                        lookup_options,
                                                        client_revision)
        self.assertEqual(status[0], NT_STATUS_INVALID_SID)
        self.assertEqual(count, 0)
        self.assertIsNotNone(domains)
        self.assertEqual(domains.count, 0)
        self.assertIsNotNone(names)
        self.assertEqual(names.count, 4)
        self.assertEqual(names.names[0].sid_type, lsa.SID_NAME_UNKNOWN)
        self.assertIsNone(names.names[0].name.string)
        self.assertEqual(names.names[0].sid_index, 0xffffffff)
        self.assertEqual(names.names[1].sid_type, lsa.SID_NAME_UNKNOWN)
        self.assertIsNone(names.names[1].name.string)
        self.assertEqual(names.names[1].sid_index, 0xffffffff)
        self.assertEqual(names.names[2].sid_type, lsa.SID_NAME_UNKNOWN)
        self.assertIsNone(names.names[2].name.string)
        self.assertEqual(names.names[2].sid_index, 0xffffffff)
        self.assertEqual(names.names[3].sid_type, lsa.SID_NAME_UNKNOWN)
        self.assertIsNone(names.names[3].name.string)
        self.assertEqual(names.names[3].sid_index, 0xffffffff)

        return



# user management
#
# user setpassword command
#
# Copyright Jelmer Vernooij 2010 <jelmer@samba.org>
# Copyright Theresa Halloran 2011 <theresahalloran@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from getpass import getpass

import samba.getopt as options
from samba import dsdb, generate_random_password, gensec, ldb
from samba.auth import system_session
from samba.netcmd import Command, CommandError, Option
from samba.samdb import SamDB


class cmd_user_setpassword(Command):
    """Set or reset the password of a user account.

This command sets or resets the logon password for a user account.  The username specified on the command is the sAMAccountName.  The username may also be specified using the --filter option.

If the password is not specified on the command through the --newpassword parameter, the user is prompted for the password to be entered through the command line.

It is good security practice for the administrator to use the --must-change-at-next-login option which requires that when the user logs on to the account for the first time following the password change, he/she must change the password.

The command may be run from the root userid or another authorized userid.  The -H or --URL= option can be used to execute the command against a remote server.

Example1:
samba-tool user setpassword TestUser1 --newpassword=passw0rd --URL=ldap://samba.samdom.example.com -Uadministrator%passw1rd

Example1 shows how to set the password of user TestUser1 on a remote LDAP server.  The --URL parameter is used to specify the remote target server.  The -U option is used to pass the username and password of a user that exists on the remote server and is authorized to update the server.

Example2:
sudo samba-tool user setpassword TestUser2 --newpassword=passw0rd --must-change-at-next-login

Example2 shows how an administrator would reset the TestUser2 user's password to passw0rd.  The user is running under the root userid using the sudo command.  In this example the user TestUser2 must change their password the next time they logon to the account.

Example3:
samba-tool user setpassword --filter=samaccountname=TestUser3 --newpassword=passw0rd

Example3 shows how an administrator would reset TestUser3 user's password to passw0rd using the --filter= option to specify the username.

"""
    synopsis = "%prog (<username>|--filter <filter>) [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "versionopts": options.VersionOptions,
        "credopts": options.CredentialsOptions,
    }

    takes_options = [
        Option("-H", "--URL", help="LDB URL for database or target server", type=str,
               metavar="URL", dest="H"),
        Option("--filter", help="LDAP Filter to set password on", type=str),
        Option("--newpassword", help="Set password", type=str),
        Option("--must-change-at-next-login",
               help="Force password to be changed on next login",
               action="store_true"),
        Option("--random-password",
               help="Generate random password",
               action="store_true"),
        Option("--smartcard-required",
               help="Require a smartcard for interactive logons",
               action="store_true"),
        Option("--clear-smartcard-required",
               help="Don't require a smartcard for interactive logons",
               action="store_true"),
    ]

    takes_args = ["username?"]

    def run(self, username=None, filter=None, credopts=None, sambaopts=None,
            versionopts=None, H=None, newpassword=None,
            must_change_at_next_login=False, random_password=False,
            smartcard_required=False, clear_smartcard_required=False):
        if filter is None and username is None:
            raise CommandError("Either the username or '--filter' must be specified!")

        password = newpassword

        if smartcard_required:
            if password is not None and password != '':
                raise CommandError('It is not allowed to specify '
                                   '--newpassword '
                                   'together with --smartcard-required.')
            if must_change_at_next_login:
                raise CommandError('It is not allowed to specify '
                                   '--must-change-at-next-login '
                                   'together with --smartcard-required.')
            if clear_smartcard_required:
                raise CommandError('It is not allowed to specify '
                                   '--clear-smartcard-required '
                                   'together with --smartcard-required.')

        if random_password and not smartcard_required:
            password = generate_random_password(128, 255)

        while True:
            if smartcard_required:
                break
            if password is not None and password != '':
                break
            password = getpass("New Password: ")
            passwordverify = getpass("Retype Password: ")
            if not password == passwordverify:
                password = None
                self.outf.write("Sorry, passwords do not match.\n")

        if filter is None:
            filter = "(&(objectClass=user)(sAMAccountName=%s))" % (ldb.binary_encode(username))

        lp = sambaopts.get_loadparm()
        creds = credopts.get_credentials(lp)

        creds.set_gensec_features(creds.get_gensec_features() | gensec.FEATURE_SEAL)

        samdb = SamDB(url=H, session_info=system_session(),
                      credentials=creds, lp=lp)

        if smartcard_required:
            command = ""
            try:
                command = "Failed to set UF_SMARTCARD_REQUIRED for user '%s'" % (username or filter)
                flags = dsdb.UF_SMARTCARD_REQUIRED
                samdb.toggle_userAccountFlags(filter, flags, on=True)
                command = "Failed to enable account for user '%s'" % (username or filter)
                samdb.enable_account(filter)
            except Exception as msg:
                # FIXME: catch more specific exception
                raise CommandError("%s: %s" % (command, msg))
            self.outf.write("Added UF_SMARTCARD_REQUIRED OK\n")
        else:
            command = ""
            try:
                if clear_smartcard_required:
                    command = "Failed to remove UF_SMARTCARD_REQUIRED for user '%s'" % (username or filter)
                    flags = dsdb.UF_SMARTCARD_REQUIRED
                    samdb.toggle_userAccountFlags(filter, flags, on=False)
                command = "Failed to set password for user '%s'" % (username or filter)
                samdb.setpassword(filter, password,
                                  force_change_at_next_login=must_change_at_next_login,
                                  username=username)
            except Exception as msg:
                # FIXME: catch more specific exception
                raise CommandError("%s: %s" % (command, msg))
            self.outf.write("Changed password OK\n")



# Unix SMB/CIFS implementation.
# Copyright © Catalyst IT 2023
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for Conditional ACEs, claims, and security tokens."""

import random
from samba.dcerpc import security
from samba.security import access_check
from samba.tests.token_factory import token as Token
from samba.tests.token_factory import list_to_claim
from samba.dcerpc.security import CLAIM_SECURITY_ATTRIBUTE_RELATIVE_V1
from samba.tests import TestCase, DynamicTestCase, get_env_dir
from samba.colour import c_RED
import os
from samba import NTSTATUSError
from samba.ntstatus import NT_STATUS_ACCESS_DENIED

DEFAULT_ACCESS = security.SEC_FILE_ALL
DEFAULT_ACCESS2 = (security.SEC_STD_READ_CONTROL |
                   security.SEC_ADS_LIST |
                   security.SEC_ADS_READ_PROP)


def write_c_test_on_failure(f):
    """This is a function decorator that writes a function for
    /libcli/security/tests/test_run_conditional_ace.c that runs the
    equivalent test. Why?! Because iterating over a test to debug the
    failure is slower in Python tests, but adding new tests is faster
    in Python. So the flow goes like this:

    1. add python tests, run them
    2. if nothing fails, goto 1
    3. copy the test_something() text into test_run_conditional_ace.c,
       rename it, and add it to main().
    4. `make bin/test_run_conditional_ace && rr bin/test_run_conditional_ace`
    5. `rr replay`

    and you're away. You can also just work from the Python, but a few
    runs of `make -j` after touching something in libcli/security will
    make you see why this exists.

    You might be thinking that this surely took longer to write than
    waiting 100 times for a 30 second compile, but that misses the
    point that debugging needs to be ergonomic and fun.
    """
    from json import dumps as q  # JSON quoting is C quoting, more or less

    def wrapper(name, token, sddl, access_desired):
        try:
            f(name, token, sddl, access_desired)
        except Exception:
            print()
            print('static void test_something(void **state)')
            print('{')
            print('\tINIT();')
            for s in ('sids', 'device_sids'):
                if s in token:
                    macro = ('user_sids' if s == 'sids' else s).upper()
                    v = ', '.join(q(x) for x in token[s])
                    print(f'\t{macro}({v});')
            for s in ('user_claims', 'device_claims'):
                if s in token:
                    macro = s.upper()
                    for name, values in token[s].items():
                        if isinstance(values,
                                      CLAIM_SECURITY_ATTRIBUTE_RELATIVE_V1):
                            v = '...'
                        else:
                            if not isinstance(values, (list, tuple)):
                                values = [values]
                            v = ', '.join(q(x) for x in values)
                            v = q(f"{v}")
                        print(f'\t{macro}({q(name)}, {v});')
            print(f'\tSD({q(sddl)});')
            if 'allow' in f.__name__:
                print(f'\tALLOW_CHECK({access_desired:#x});')
            else:
                print(f'\tDENY_CHECK({access_desired:#x});')
            print('}')
            print()
            raise
    return wrapper


class ConditionalAceClaimsBase(TestCase):
    maxDiff = 0

    @classmethod
    def setUpDynamicTestCases(cls):
        cls.domain_sid = security.dom_sid("S-1-22-333-4444")
        seen = set()

        for i, row in enumerate(cls.data):
            token, sddl, access_desired = row
            name = f'{i+1:03}-{token}-{sddl}-{access_desired}'
            if len(name) > 150:
                name = f"{name[:125]}+{len(name) - 125}-more-characters"

            if name in seen:
                print(f"seen {row} after {len(seen)}")
            seen.add(name)

            if cls.allow:
                cls.generate_dynamic_test('test_allow',
                                          name, token, sddl, access_desired)
            else:
                cls.generate_dynamic_test('test_deny',
                                          name, token, sddl, access_desired)

        fuzz_seed_dir = get_env_dir('SAMBA_WRITE_FUZZ_STRINGS_DIR')
        if fuzz_seed_dir is not None:
            cls._write_sddl_strings_for_fuzz_seeds(fuzz_seed_dir)

    @classmethod
    def _write_sddl_strings_for_fuzz_seeds(cls, fuzz_seed_dir):
        """write all the SDDL strings we have into a directory as individual
        files, using a naming convention beloved of fuzzing engines.

        To run this set an environment variable; see
        cls.setUpDynamicTestCases(), below.

        Note this will only run in subclasses annotated with @DynamicTestCase.
        """
        from hashlib import md5
        for _, sddl, _ in cls.data:
            name = md5(sddl.encode()).hexdigest()
            with open(os.path.join(fuzz_seed_dir, name), 'w') as f:
                f.write(sddl)

    @write_c_test_on_failure
    def _test_allow_with_args(self, _token, sddl, access_desired):
        if isinstance(_token, dict):
            token = Token(**_token)
        else:
            token = _token
        sd = security.descriptor.from_sddl(sddl, self.domain_sid)
        try:
            granted = access_check(sd, token, access_desired)
        except NTSTATUSError as e:
            print(c_RED(sddl))
            print(c_RED(_token))
            if e.args[0] != NT_STATUS_ACCESS_DENIED:
                raise
            self.fail("access was denied")

        self.assertEqual(granted, access_desired)

    @write_c_test_on_failure
    def _test_deny_with_args(self, token, sddl, access_desired):
        if isinstance(token, dict):
            token = Token(**token)
        sd = security.descriptor.from_sddl(sddl, self.domain_sid)
        try:
            granted = access_check(sd, token, access_desired)
        except NTSTATUSError as e:
            if e.args[0] == NT_STATUS_ACCESS_DENIED:
                return
            self.fail(f"failed with {e}, not access denied")

        self.fail("access allowed")


@DynamicTestCase
class AllowTests(ConditionalAceClaimsBase):
    name = "allow"
    allow = True
    data = [
        (  # device_claims
            {'sids': ['WD', 'AA'],
             'device_claims': {"colour":["orange", "blue"]}},
            ('D:(XA;;0x1f;;;AA;'
             '(@Device.colour == {"orange", "blue"}))'),
            0x10),
        (  # device_claims, int >=
            {'sids': ['WD', 'AA'],
             'device_claims': {"legs": 4}},
            ('D:(XA;;0x1f;;;AA;(@Device.legs >= 1))'),
            0x10),
        (  # device_claims, int
            {'sids': ['WD', 'AA'],
             'device_claims': {"legs": 1}},
            ('D:(XA;;0x1f;;;AA;(@Device.legs == 1))'),
            0x10),
        (  # device_member_of && member_of
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            ("D:(XA;;0x1f;;;AA;"
             "(Device_Member_of{SID(BA)} && Member_of{SID(WD)}))"),
            0x10),
        (  # device_member_of || member_of, both true
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            ("D:(XA;;0x1f;;;AA;"
             "(Device_Member_of{SID(AA)} || Member_of{SID(WD)}))"),
            0x10),
        (  # device_member_of || member_of, second true
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            ("D:(XA;;0x1f;;;AA;"
             "(Device_Member_of{SID(AA)} || Member_of{SID(WD)}))"),
            0x10),
        (  # device_member_of || member_of, first true
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            ("D:(XA;;0x1f;;;AA;"
             "(Device_Member_of{SID(BG)} || Member_of{SID(WR)}))"),
            0x10),
        (  # single SID, Member_of_Any
            {'sids': ['S-1-222-333']},
            ("D:(XA;;0x1ff;;;S-1-222-333;(Member_of_Any{SID(S-1-222-333)}))"),
            0x1),
        ({'sids': ['S-1-1-0']}, "O:S-1-1-0D:(A;;0x1ff;;;WD)", DEFAULT_ACCESS),
        ({'sids': ['S-1-1-0']},
         "O:S-1-1-0D:(XA;;0x1ff;;;WD;(Member_of{SID(S-1-1-0)}))",
         DEFAULT_ACCESS),
        ({'sids': ['S-1-1-0', 'S-1-222-333']},
         "O:S-1-1-0D:(XA;;0x1ff;;;WD;(Member_of{SID(S-1-1-0)}))",
         DEFAULT_ACCESS),
        ({'sids': ['WD', 'S-1-222-333']},
         "O:S-1-1-0D:(XA;;0x1ff;;;WD;(Member_of{SID(S-1-1-0)}))",
         DEFAULT_ACCESS),
        (  # a single SID, not a composite
            {'sids': ['S-1-1-0', 'S-1-222-333']},
            "O:S-1-1-0D:(XA;;0x1ff;;;WD;(Member_of SID(S-1-1-0)))",
            DEFAULT_ACCESS),
        (  # a single SID, not a composite, without space after Member_of
            {'sids': ['S-1-1-0', 'S-1-222-333']},
            "O:S-1-1-0D:(XA;;0x1ff;;;WD;(Member_of\nSID(S-1-1-0)))",
            DEFAULT_ACCESS),
        (  # a single SID, not a composite, Member_of_Any
            {'sids': ['S-1-1-0', 'S-1-222-333']},
            "O:S-1-1-0D:(XA;;0x1ff;;;WD;(Member_of_Any SID(S-1-1-0)))",
            DEFAULT_ACCESS),
        (  # Member_of_Any
            {'sids': ['S-1-1-0', 'S-1-222-333']},
            "O:S-1-1-0D:(XA;;0x1;;;WD;(Member_of_Any{SID(AS),SID(WD)}))",
            0x1),
        ({'sids': ['S-1-1-0', 'S-1-222-333']},
         ("O:S-1-1-0D:"
          "(XA;;0x1ff;;;WD;(Member_of_Any{SID(S-1-1-0), SID(S-1-222-333)}))"),
         DEFAULT_ACCESS),
        ({'sids': ['S-1-1-0', 'S-1-222-333']},
         ("O:S-1-1-0D:"
          "(XA;;0x1ff;;;WD;(Member_of_Any{SID(S-1-1-334), SID(S-1-222-333)}))"),
         DEFAULT_ACCESS),
        ({'sids': ['S-1-1-0', 'S-1-222-333']},
         ("D:(XA;;0x1ff;;;WD;(Member_of_Any{SID(S-1-222-333)}))"),
         DEFAULT_ACCESS),
        ({'sids': ['S-1-77-88-99', 'AA']},
         "D:(XA;;0x1f;;;AA;(Member_of{SID(S-1-77-88-99)}))",
         0x10),
        (  # device_member_of
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            "D:(XA;;0x1f;;;AA;(Device_Member_of{SID(BA)}))",
            0x10),
        (  # device_member_of
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            "D:(XA;;0x1f;;;AA;(Device_Member_of{SID(BA)}))",
            0x10),
        (  # not (!) member_of
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            "D:(XA;;0x1f;;;AA;(! (Member_of{SID(BA)})))",
            0x10),
        (  # not not (!!) member_of
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            "D:(XA;;0x1f;;;AA;(!(! (Member_of{SID(AA)}))))",
            0x10),
        (  # not * 8 (!!!! !!!!) member_of
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            "D:(XA;;0x1f;;;AA;(!(!(!(!(!(!(!(!( Member_of{SID(AA)}))))))))))",
            0x10),
        (  # not * 9 (!!! !!! !!!) member_of
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            "D:(XA;;0x1f;;;AA;(!(!(!( !(!(!( !(!(!(Member_of{SID(BA)})))))))))))",
            0x10),
        (  # not * 9 (!!! !!! !!!) Not_Member_of
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            ("D:(XA;;0x1f;;;AA;"
             "(!(!(!(  !(!(!(  !(!(!(  Not_Member_of{SID(AA)})))))))))))"),
            0x10),
        (  #resource ACE
            {'sids': ['WD', 'AA'],
             'device_claims': {"colour": ["blue"]}},
            ('D:(XA;;0x1f;;;AA;(@Device.colour Contains @Resource.colour))'
             'S:(RA;;;;;WD;("colour",TS,0,"blue"))'),
            0x10),
        (  #resource ACE ==
            {'sids': ['WD', 'AA'],
             'device_claims': {"colour": ["blue"]}},
            ('D:(XA;;0x1f;;;AA;(@Device.colour == @Resource.colour))'
             'S:(RA;;;;;WD;("colour",TS,0,"blue"))'),
            0x10),
        (  # device_claims, comparing single to single
            {'sids': ['WD', 'AA'],
             'device_claims': {"colour": "blue"}},
            ('D:(XA;;0x1f;;;AA;(@Device.colour == "blue"))'),
            0x10),
        (  # device_claims == user_claims
            {'sids': ['WD', 'AA'],
             'user_claims': {"colour": "blue"},
             'device_claims': {"colour": "blue"}},
            ('D:(XA;;0x1f;;;AA;(@User.colour == @Device.colour))'),
            0x10),
        (  #resource ACE multi
            {'sids': ['WD', 'AA'],
             'device_claims': {"colour": ["blue", "red"]}},
            ('D:(XA;;0x1f;;;AA;(@Device.colour Contains @Resource.colour))'
             'S:(RA;;;;;WD;("colour",TS,0,"blue", "red"))'),
            0x10),
    ]


@DynamicTestCase
class DenyTests(ConditionalAceClaimsBase):
    name = "allow"
    allow = False
    data = [
        ({}, "", DEFAULT_ACCESS),
        ({'sids': ['S-1-1-0']}, "O:S-1-1-0D:(A;;0x1fe;;;WD)", DEFAULT_ACCESS),
        ({}, "O:WDD:(A;;GACR;;;CO)", DEFAULT_ACCESS),
        ({'sids': ['S-1-1-0', 'S-1-222-444']},
         ("D:(XA;;0x1ff;;;WD;(Member_of_Any{SID(S-1-222-333)}))"),
         0x1),
        (  # Without explicit 'everyone' SID in list of SIDs, this is
           # denied because the ACE SID 'WD' doesn't match.
            {'sids': ['S-1-222-333']},
         ("D:(XA;;0x1ff;;;WD;(Member_of_Any{SID(S-1-222-333)}))"),
         0x1),
        (  # device_member_of && member_of, both false
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            ("D:(XA;;0x1f;;;AA;"
             "(Device_Member_of{SID(AA)} && Member_of{SID(WR)}))"),
            0x10),
        (  # device_member_of && member_of, first false
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            ("D:(XA;;0x1f;;;AA;"
             "(Device_Member_of{SID(AA)} && Member_of{SID(WD)}))"),
            0x10),
        (  # device_member_of && member_of, second false
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            ("D:(XA;;0x1f;;;AA;"
             "(Device_Member_of{SID(BA)} && Member_of{SID(BA)}))"),
            0x10),
        (  # device_member_of || member_of, both false
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            ("D:(XA;;0x1f;;;AA;"
             "(Device_Member_of{SID(AA)} || Member_of{SID(WR)}))"),
            0x10),
        (  # device_claims, comparing composite to single
            {'sids': ['WD', 'AA'],
             'device_claims': {"colour": ["orange", "blue"]}},
            ('D:(XA;;0x1f;;;AA;(@Device.colour == "blue"))'),
            0x10),
        (  # not (!) member_of
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            "D:(XA;;0x1f;;;AA;(! (Member_of{SID(AA)})))",
            0x10),
        (  # not not (!!) member_of
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            "D:(XA;;0x1f;;;AA;(!(!( Member_of{SID(BA)}))))",
            0x10),
        (  # not * 8 (!!!! !!!!) member_of
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            "D:(XA;;0x1f;;;AA;(!(!( !(!( !(!( !(!(Member_of{SID(BA)}))))))))))",
            0x10),
        (  # not * 3 (!!!) member_of
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            "D:(XA;;0x1f;;;AA;(!(!(!(Member_of{SID(AA)})))))",
            0x10),
        (  # not * 3 (!!!) Not_Member_of
            {'sids': ['WD', 'AA'],
             'device_sids': ['BA', 'BG']},
            "D:(XA;;0x1f;;;AA;(!(!(!(Not_Member_of{SID(BA)})))))",
            0x10),
    ]


def _int_range(n, n_dupes=0, random_seed=None):
    """Makes a list of stringified integers.

    If n_unique is specified and less than n, there will be that many unique
    values (and hence some duplicates). If random_seed is set, the list will be
    shuffled.
    """
    claims = [str(x) for x in range(n)]

    if random_seed is None:
        if n_dupes:
            claims *= 1 + (n + n_dupes) // n
        return claims[:n + n_dupes]

    random.seed(random_seed)
    for i in range(n_dupes):
        # this purposefully skews the distribution.
        claims.append(random.choice(claims))

    random.shuffle(claims)
    return claims


def _str_range(n, n_dupes=0, random_seed=None, mix_case=False):
    """Create a list of strings with somewhat controllable disorder.
    """
    ints = _int_range(n, n_dupes, random_seed)
    claims = [f'a{i}' for i in ints]

    if mix_case:
        if random_seed is None:
            random.seed(0)
        for i in range(len(claims)):
            if random.random() < 0.5:
                claims[i] = claims[i].upper()

    return claims


def claim_str_range(*args, name="foo", case_sensitive=False, **kwargs):
    """String value range as a CLAIM_SECURITY_ATTRIBUTE_RELATIVE_V1"""
    vals = _str_range(*args, **kwargs)
    claim = list_to_claim(name, vals, case_sensitive=case_sensitive)
    return claim


def claim_int_range(*args, name="foo", case_sensitive=False, **kwargs):
    """Int value range as a CLAIM_SECURITY_ATTRIBUTE_RELATIVE_V1"""
    vals = _int_range(*args, **kwargs)
    claim = list_to_claim(name, vals, case_sensitive=case_sensitive)
    return claim


def ra_str_range(*args, name="foo", case_sensitive=False, **kwargs):
    """Make a string claim as a resource attribute"""
    claim = _str_range(*args, **kwargs)
    values = '","'.join(claim)
    c = (2 if case_sensitive else 0)
    return f'(RA;;;;;WD;("{name}",TS,{c},"{values}"))'


def ra_int_range(*args, name="foo", unsigned=False, **kwargs):
    """Return an integer claim range as a resource attribute."""
    ints = _int_range(*args, **kwargs)
    values = ','.join(str(x) for x in ints)
    return f'(RA;;;;;WD;("{name}",T{"U" if unsigned else "I"},0,{values}))'


def composite_int(*args, **kwargs):
    """Integer conditional ACE composite"""
    claim = _int_range(*args, **kwargs)
    values = ', '.join(claim)
    return '{' + values + '}'


def composite_str(*args, **kwargs):
    """String conditional ACE composite"""
    claim = _str_range(*args, **kwargs)
    values = '", "'.join(claim)
    return '{"' + values + '"}'


@DynamicTestCase
class ConditionalAceLargeComposites(ConditionalAceClaimsBase):
    """Here we are dynamically generating claims and composites with large numbers
    of members, and using them in comparisons. Sometimes the comparisons are
    meant to fail, and sometimes not.
    """
    maxDiff = 0

    @classmethod
    def setUpDynamicTestCases(cls):
        cls.domain_sid = security.dom_sid("S-1-22-333-4444")
        for i, row in enumerate(cls.data):
            name, allow, token, sddl = row
            name = f'{i+1:03}-{name}'
            if 'sids' not in token:
                token['sids'] = ['AU', 'WD']
            if allow:
                cls.generate_dynamic_test('test_allow',
                                          name, token, sddl, 0x10)
            else:
                cls.generate_dynamic_test('test_deny',
                                          name, token, sddl, 0x10)

        fuzz_seed_dir = get_env_dir('SAMBA_WRITE_FUZZ_STRINGS_DIR')
        if fuzz_seed_dir is not None:
            cls._write_sddl_strings_for_fuzz_seeds(fuzz_seed_dir)


    data = [
        (
            "90-disorderly-strings-claim-vs-claim-case-sensitive-with-dupes",
            False,
            {'user_claims': {"c": claim_str_range(90,
                                                  random_seed=2),
                             "d": claim_str_range(90, 90,
                                                  case_sensitive=True,
                                                  random_seed=3)}},
            ('D:(XA;;FA;;;WD;(@USER.c == @USER.d))')
        ),
        (
            # this one currently fails before we get to compare_composites()
            "0-vs-0",
            True,
            {'user_claims': {"c": claim_str_range(0)}},
            ('D:(XA;;FA;;;WD;(@USER.c == @USER.c))')
        ),
        (
            "50-orderly-strings",
            True,
            {'user_claims': {"c": claim_str_range(50)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(50)}))')
        ),
        (
            "50-disorderly-strings-same-disorder",
            True,
            {'user_claims': {"c": claim_str_range(50, random_seed=1)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(50, random_seed=1)}))')
        ),
        (
            "200-disorderly-strings",
            True,
            {'user_claims': {"c": claim_str_range(200, random_seed=1)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(200, random_seed=2)}))')
        ),
        (
            "50-orderly-vs-disorderly-strings",
            True,
            {'user_claims': {"c": claim_str_range(50)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(50, random_seed=1)}))')
        ),
        (
            "50-disorderly-vs-orderly-strings",
            True,
            {'user_claims': {"c": claim_str_range(50, random_seed=1)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(50)}))')
        ),
        (
            "99-orderly-strings",
            True,
            {'user_claims': {"c": claim_str_range(99)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(99)}))')
        ),
        (
            "99-disorderly-strings",
            True,
            {'user_claims': {"c": claim_str_range(99, random_seed=1)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(99, random_seed=2)}))')
        ),
        (
            "99-orderly-vs-disorderly-strings",
            True,
            {'user_claims': {"c": claim_str_range(99)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(99, random_seed=1)}))')
        ),
        (
            "99-disorderly-vs-orderly-strings",
            True,
            {'user_claims': {"c": claim_str_range(99, random_seed=1)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(99)}))')
        ),
        (
            "39-orderly-strings-vs-39+60-dupes",
            True,
            {'user_claims': {"c": claim_str_range(39)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(39, 60)}))')
        ),
        (
            "39-disorderly-strings-vs-39+60-dupes",
            True,
            {'user_claims': {"c": claim_str_range(39, random_seed=1)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(39, 60, random_seed=1)}))')
        ),
        (
            "39-orderly-vs-disorderly-strings-vs-39+60-dupes",
            True,
            {'user_claims': {"c": claim_str_range(39)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(39, 60, random_seed=1)}))')
        ),
        (
            "39-disorderly-vs-orderly-strings-vs-39+60-dupes",
            True,
            {'user_claims': {"c": claim_str_range(39, random_seed=1)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(39, 60)}))')
        ),
        (
            "3-orderly-strings-vs-3+60-dupes",
            True,
            {'user_claims': {"c": claim_str_range(3)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(3, 60)}))')
        ),
        (
            "3-disorderly-strings-vs-3+60-dupes",
            True,
            {'user_claims': {"c": claim_str_range(3, random_seed=1)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(3, 60, random_seed=1)}))')
        ),
        (
            "3-orderly-vs-disorderly-strings-vs-3+60-dupes",
            True,
            {'user_claims': {"c": claim_str_range(3)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(3, 60, random_seed=1)}))')
        ),
        (
            "3-disorderly-vs-orderly-strings-vs-3+60-dupes",
            True,
            {'user_claims': {"c": claim_str_range(3, random_seed=1)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(3, 60)}))')
        ),
        (
            "3-orderly-strings-vs-3+61-dupes",
            True,
            {'user_claims': {"c": claim_str_range(3)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(3, 61)}))')
        ),

        (
            "63-orderly-strings-vs-62+1-dupe",
            False,
            {'user_claims': {"c": claim_str_range(63)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(62, 1)}))')
        ),
        (
            "102+1-dupe-vs-102+1-dupe",
            False,
            # this is an invalid claim
            {'user_claims': {"c": claim_str_range(102, 1)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(102, 1)}))')
        ),
        (
            "0-vs-1",
            False,
            {'user_claims': {"c": claim_str_range(0),
                             "d": claim_str_range(1)}},
            ('D:(XA;;FA;;;WD;(@USER.c == @USER.d))')
        ),
        (
            "2+1-dupe-vs-2+1-dupe",
            False,
            {'user_claims': {"c": claim_str_range(2, 1)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(2, 1)}))')
        ),
        (
            "63-disorderly-strings-vs-62+1-dupe",
            False,
            {'user_claims': {"c": claim_str_range(63, random_seed=1)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(62, 1, random_seed=1)}))')
        ),
        (
            "63-disorderly-strings-vs-63+800-dupe",
            True,
            {'user_claims': {"c": claim_str_range(63, random_seed=1)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(63, 800, random_seed=1)}))')
        ),
        (
            "63-disorderly-strings-vs-62+800-dupe",
            False,
            {'user_claims': {"c": claim_str_range(63, random_seed=1)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(62, 800, random_seed=1)}))')
        ),
        (
            "9-orderly-strings",
            True,
            {'user_claims': {"c": claim_str_range(9)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(9)}))')
        ),
        (
            "9-orderly-strings-claim-vs-itself",
            True,
            {'user_claims': {"c": claim_str_range(9)}},
            ('D:(XA;;FA;;;WD;(@USER.c == @USER.c))')
        ),
        (
            "300-orderly-strings-claim-vs-itself",
            True,
            {'user_claims': {"c": claim_str_range(300)}},
            ('D:(XA;;FA;;;WD;(@USER.c == @USER.c))')
        ),
        (
            "900-disorderly-strings-claim-vs-claim",
            True,
            {'user_claims': {"c": claim_str_range(900, random_seed=1),
                             "d": claim_str_range(900, random_seed=1)}},
            ('D:(XA;;FA;;;WD;(@USER.c == @USER.d))')
        ),
        (
            "9-orderly-strings-claim-mixed-case-vs-claim-case-sensitive",
            False,
            {'user_claims': {"c": claim_str_range(9, mix_case=True),
                             "d": claim_str_range(9, case_sensitive=True)}},
            ('D:(XA;;FA;;;WD;(@USER.c == @USER.d))')
        ),
        (
            "9-disorderly-strings-claim-vs-claim-case-sensitive-mixed-case",
            False,
            {'user_claims': {"c": claim_str_range(9,random_seed=1),
                             "d": claim_str_range(9,
                                                  mix_case=True,
                                                  case_sensitive=True)}},
            ('D:(XA;;FA;;;WD;(@USER.c == @USER.d))')
        ),
        (
            "9-disorderly-strings-claim-vs-claim-case-sensitive-both-mixed-case",
            False,
            {'user_claims': {"c": claim_str_range(9,
                                                  mix_case=True,
                                                  random_seed=1),
                             "d": claim_str_range(9,
                                                  mix_case=True,
                                                  case_sensitive=True)}},
            ('D:(XA;;FA;;;WD;(@USER.c == @USER.d))')
        ),
        (
            "9-disorderly-strings-claim-vs-claim-case-sensitive-ne",
            True,
            {'user_claims': {"c": claim_str_range(9,random_seed=1),
                             "d": claim_str_range(9,
                                                  mix_case=True,
                                                  case_sensitive=True)}},
            ('D:(XA;;FA;;;WD;(@USER.c != @USER.d))')
        ),

        (
            "5-disorderly-strings-claim-vs-claim-case-sensitive-with-dupes-all-mixed-case",
            False,
            {'user_claims': {"c": claim_str_range(5,
                                                  mix_case=True,
                                                  random_seed=2),
                             "d": claim_str_range(5, 5,
                                                  mix_case=True,
                                                  random_seed=3,
                                                  case_sensitive=True)}},
            ('D:(XA;;FA;;;WD;(@USER.c == @USER.d))')
        ),
        (
            "90-disorderly-strings-claim-vs-int-claim",
            False,
            {'user_claims': {"c": claim_str_range(90,
                                                  random_seed=2),
                             "d": claim_int_range(90,
                                                  random_seed=3)}},
            ('D:(XA;;FA;;;WD;(@USER.c == @USER.d))')
        ),
        (
            "90-disorderly-ints-claim-vs-string-claim",
            False,
            {'user_claims': {"c": claim_int_range(90,
                                                  random_seed=2),
                             "d": claim_str_range(90,
                                                  random_seed=3)}},
            ('D:(XA;;FA;;;WD;(@USER.c == @USER.d))')
        ),
        (
            "9-disorderly-strings-vs-9+90-dupes",
            True,
            {'user_claims': {"c": claim_str_range(9, random_seed=1)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(9, 90, random_seed=1)}))')
        ),
        (
            "9-disorderly-strings-vs-9+90-dupes-case-sensitive",
            True,
            {'user_claims': {"c": claim_str_range(9, random_seed=1, case_sensitive=True)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(9, 90, random_seed=2)}))')
        ),
        (
            "9-disorderly-strings-vs-9+90-dupes-mixed-case",
            True,
            {'user_claims': {"c": claim_str_range(9, random_seed=1, mix_case=True)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(9, 90, random_seed=2, mix_case=True)}))')
        ),
        (
            "9-disorderly-strings-vs-9+90-dupes-mixed-case-case-sensitive",
            False,
            {'user_claims': {"c": claim_str_range(9, random_seed=1, mix_case=True,
                                                  case_sensitive=True)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(9, 90, random_seed=2, mix_case=True)}))')
        ),
        (
            "99-disorderly-strings-vs-9+90-dupes-mixed-case",
            False,
            {'user_claims': {"c": claim_str_range(99, random_seed=1, mix_case=True)}},
            (f'D:(XA;;FA;;;WD;(@USER.c == {composite_str(9, 90, random_seed=2, mix_case=True)}))')
        ),

        (
            "RA-99-disorderly-strings-vs-9+90-dupes-mixed-case",
            False,
            {},
            ('D:(XA;;FA;;;WD;(@RESOURCE.c == '
             f'{composite_str(9, 90, random_seed=1, mix_case=True)}))'
             f'S:{ra_str_range(99, random_seed=2, mix_case=True)}'
             )
        ),
        (
            "RA-9+90-dupes-disorderly-strings-vs-9+90-dupes-mixed-case",
            False,
            {},
            ('D:(XA;;FA;;;WD;(@RESOURCE.c == '
             f'{composite_str(9, 90, random_seed=1, mix_case=True)}))'
             f'S:{ra_str_range(9, 90, random_seed=2, mix_case=True)}'
             )
        ),
        (
            "90-disorderly-strings-claim-vs-missing-claim",
            False,
            {'user_claims': {"c": claim_str_range(90,
                                                  random_seed=2)}},
            ('D:(XA;;FA;;;WD;(@USER.c == @USER.d))')
        ),
        (
            "missing-claim-vs-90-disorderly-strings",
            False,
            {'user_claims': {"c": claim_str_range(90,
                                                  random_seed=2)}},
            ('D:(XA;;FA;;;WD;(@USER.z == @USER.c))')
        ),

        (
            "RA-9-disorderly-strings-vs-9-mixed-case",
            False,
            {'user_claims': {"c": claim_str_range(9,
                                                  random_seed=1,
                                                  mix_case=True),
                             }
             },
            ('D:(XA;;FA;;;WD;(@RESOURCE.c == @User.c))'
             f'S:{ra_str_range(9, random_seed=2, mix_case=True)}'
             )
        ),

        (
            "9-disorderly-strings-vs-9-RA-mixed-case",
            False,
            {'user_claims': {"c": claim_str_range(9,
                                                  random_seed=1,
                                                  mix_case=True),
                             }
             },
            ('D:(XA;;FA;;;WD;(@user.c == @resource.c))'
             f'S:{ra_str_range(9, random_seed=2, mix_case=True)}'
             )
        ),

        (
            "RA-29-disorderly-strings-vs-29-mixed-case",
            False,
            {'user_claims': {"c": claim_str_range(29,
                                                  random_seed=1,
                                                  mix_case=True),
                             }
             },
            ('D:(XA;;FA;;;WD;(@RESOURCE.c == @User.c))'
             f'S:{ra_str_range(29, random_seed=2, mix_case=True)}'
             )
        ),
        (
            "0-vs-0-ne",
            False,
            {'user_claims': {"c": claim_str_range(0)}},
            ('D:(XA;;FA;;;WD;(@USER.c != @USER.c))')
        ),
        (
            "1-vs-1",
            True,
            {'user_claims': {"c": claim_str_range(1)}},
            ('D:(XA;;FA;;;WD;(@USER.c == @USER.c))')
        ),
        (
            "1-vs-1-ne",
            False,
            {'user_claims': {"c": claim_str_range(1)}},
            ('D:(XA;;FA;;;WD;(@USER.c != @USER.c))')
        ),
    ]
